from tensorflow.keras.layers import Layer
import tensorflow as tf

class LogSoftmax(Layer):

    """Log Softmax activation function.

    Input shape:
        Arbitrary. Use the keyword argument `input_shape`
        (tuple of integers, does not include the samples axis)
        when using this layer as the first layer in a model.

    Output shape:
        Same shape as the input.

    Arguments:
        axis: Integer, axis along which the softmax normalization is applied.
    """

    def __init__(self, axis=-1, **kwargs):
        super(LogSoftmax, self).__init__(**kwargs)
        self.supports_masking = True
        self.axis = axis

    def call(self, inputs):
        return tf.nn.log_softmax(inputs, axis=self.axis)

    def get_config(self):
        config = {'axis': self.axis}
        base_config = super(LogSoftmax, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def compute_output_shape(self, input_shape):
        return input_shape






import tensorflow as tf

import tensorflow.keras.backend as K
from tensorflow.keras.layers import Layer, InputSpec, Concatenate
from tensorflow.keras import initializers


class AttentionLayer2D(Layer):

    """
    Attention layer (2-D) from the self attention GAN

    taken from the following python implementation

    https://stackoverflow.com/questions/50819931/self-attention-gan-in-keras

    based on the following paper:

    https://arxiv.org/abs/1805.08318

    Arguments
    ---------
    number_of_channels : integer
        Number of channels

    Returns
    -------
    Layer
        A keras layer

    """

    def __init__(self, number_of_channels, **kwargs):

        super(AttentionLayer2D, self).__init__(**kwargs)

        self.number_of_channels = number_of_channels
        self.number_of_filters_f_g = self.number_of_channels // 8
        self.number_of_filters_h = self.number_of_channels

    def build(self, input_shape):

        kernel_shape_f_g = (1, 1) + (self.number_of_channels, self.number_of_filters_f_g)
        kernel_shape_h = (1, 1) + (self.number_of_channels, self.number_of_filters_h)

        self.gamma = self.add_weight(shape=[1],
                                     initializer=initializers.zeros(),
                                     trainable=True,
                                     name="gamma")
        self.kernel_f = self.add_weight(shape=kernel_shape_f_g,
                                        initializer=initializers.glorot_uniform(),
                                        trainable=True,
                                        name="kernel_f")
        self.kernel_g = self.add_weight(shape=kernel_shape_f_g,
                                        initializer=initializers.glorot_uniform(),
                                        trainable=True,
                                        name="kernel_g")
        self.kernel_h = self.add_weight(shape=kernel_shape_h,
                                        initializer=initializers.glorot_uniform(),
                                        trainable=True,
                                        name="kernel_h")
        self.bias_f = self.add_weight(shape=(self.number_of_filters_f_g,),
                                      initializer=initializers.zeros(),
                                      trainable=True,
                                      name="bias_f")
        self.bias_g = self.add_weight(shape=(self.number_of_filters_f_g,),
                                      initializer=initializers.zeros(),
                                      trainable=True,
                                      name="bias_g")
        self.bias_h = self.add_weight(shape=(self.number_of_filters_h,),
                                      initializer=initializers.zeros(),
                                      trainable=True,
                                      name="bias_h")

        super(AttentionLayer2D, self).build(input_shape)

        self.input_spec = InputSpec(ndim=4, axes={3: input_shape[-1]})
        self.built = True

    def call(self, inputs, mask=None):

        def flatten(x):
            input_shape = K.shape(x)
            output_shape = (input_shape[0], input_shape[1] * input_shape[2], input_shape[3])
            x_flat = K.reshape(x, shape=output_shape)
            return( x_flat )

        f = K.conv2d(inputs, kernel=self.kernel_f, strides=(1, 1), padding='same')
        f = K.bias_add(f, self.bias_f)
        g = K.conv2d(inputs, kernel=self.kernel_g, strides=(1, 1), padding='same')
        g = K.bias_add(g, self.bias_g)
        h = K.conv2d(inputs, kernel=self.kernel_h, strides=(1, 1), padding='same')
        h = K.bias_add(h, self.bias_h)

        f_flat = flatten(f)
        g_flat = flatten(g)
        h_flat = flatten(h)

        s = tf.matmul(g_flat, f_flat, transpose_b = True)
        beta = K.softmax(s, axis=-1)
        o = K.reshape(K.batch_dot(beta, h_flat), shape=K.shape(inputs))

        x = self.gamma * o + inputs
        return(x)

    def compute_output_shape(self, input_shape):
        return(input_shape)

    def get_config(self):
        config = {"number_of_channels": self.number_of_channels}
        base_config = super(AttentionLayer2D, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class AttentionLayer3D(Layer):

    """
    Attention layer (3-D) from the self attention GAN

    taken from the following python implementation

    https://stackoverflow.com/questions/50819931/self-attention-gan-in-keras

    based on the following paper:

    https://arxiv.org/abs/1805.08318

    Arguments
    ---------
    number_of_channels : integer
        Number of channels

    Returns
    -------
    Layer
        A keras layer

    Example
    -------
    >>> input_shape = (100, 100, 3)
    >>> input = Input(shape=input_shape)
    >>> number_of_filters = 64
    >>> outputs = Conv2D(filters=number_of_filters, kernel_size=2)(input)
    >>> outputs = AttentionLayer2D(number_of_channels=number_of_filters)(outputs)
    >>> model = Model(inputs=input, outputs=outputs)

    """

    def __init__(self, number_of_channels, **kwargs):

        super(AttentionLayer3D, self).__init__(**kwargs)

        self.number_of_channels = number_of_channels
        self.number_of_filters_f_g = self.number_of_channels // 8
        self.number_of_filters_h = self.number_of_channels

    def build(self, input_shape):

        kernel_shape_f_g = (1, 1, 1) + (self.number_of_channels, self.number_of_filters_f_g)
        kernel_shape_h = (1, 1, 1) + (self.number_of_channels, self.number_of_filters_h)

        self.gamma = self.add_weight(shape=[1],
                                     initializer=initializers.zeros(),
                                     trainable=True,
                                     name="gamma")
        self.kernel_f = self.add_weight(shape=kernel_shape_f_g,
                                        initializer=initializers.glorot_uniform(),
                                        trainable=True,
                                        name="kernel_f")
        self.kernel_g = self.add_weight(shape=kernel_shape_f_g,
                                        initializer=initializers.glorot_uniform(),
                                        trainable=True,
                                        name="kernel_g")
        self.kernel_h = self.add_weight(shape=kernel_shape_h,
                                        initializer=initializers.glorot_uniform(),
                                        trainable=True,
                                        name="kernel_h")
        self.bias_f = self.add_weight(shape=(self.number_of_filters_f_g,),
                                      initializer=initializers.zeros(),
                                      trainable=True,
                                      name="bias_f")
        self.bias_g = self.add_weight(shape=(self.number_of_filters_f_g,),
                                      initializer=initializers.zeros(),
                                      trainable=True,
                                      name="bias_g")
        self.bias_h = self.add_weight(shape=(self.number_of_filters_h,),
                                      initializer=initializers.zeros(),
                                      trainable=True,
                                      name="bias_h")

        super(AttentionLayer3D, self).build(input_shape)

        self.input_spec = InputSpec(ndim=5, axes={4: input_shape[-1]})
        self.built = True

    def call(self, inputs, mask=None):

        def flatten(x):
            input_shape = K.shape(x)
            output_shape = (input_shape[0], input_shape[1] * input_shape[2] * input_shape[3], input_shape[4])
            x_flat = K.reshape(x, shape=output_shape)
            return( x_flat )

        f = K.conv3d(inputs, kernel=self.kernel_f, strides=(1, 1, 1), padding='same')
        f = K.bias_add(f, self.bias_f)
        g = K.conv3d(inputs, kernel=self.kernel_g, strides=(1, 1, 1), padding='same')
        g = K.bias_add(g, self.bias_g)
        h = K.conv3d(inputs, kernel=self.kernel_h, strides=(1, 1, 1), padding='same')
        h = K.bias_add(h, self.bias_h)

        f_flat = flatten(f)
        g_flat = flatten(g)
        h_flat = flatten(h)

        s = tf.matmul(g_flat, f_flat, transpose_b = True)
        beta = K.softmax(s, axis=-1)
        o = K.reshape(K.batch_dot(beta, h_flat), shape=K.shape(inputs))

        x = self.gamma * o + inputs
        return(x)

    def compute_output_shape(self, input_shape):
        return(input_shape)

    def get_config(self):
        config = {"number_of_channels": self.number_of_channels}
        base_config = super(AttentionLayer3D, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))





from tensorflow.keras.models import Model
from tensorflow.keras.layers import (Input, Activation, Add, BatchNormalization,
                          Conv2D, Conv2DTranspose,
                          MaxPooling2D, UpSampling2D,
                          Conv3D, Conv3DTranspose,
                          MaxPooling3D, UpSampling3D)

def create_resnet_super_resolution_model_2d(input_image_size,
                                            convolution_kernel_size=(3, 3),
                                            number_of_filters=64,
                                            number_of_residual_blocks=5,
                                            number_of_resnet_blocks=1
                                           ):
    """
    2-D implementation of the ResNet image super resolution architecture.

    Creates a keras model of the expanded image super resolution deep learning
    framework based on the following python implementation:

            https://github.com/titu1994/Image-Super-Resolution

    Arguments
    ---------
    input_image_size : tuple of length 3
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    convolution_kernel_size : 2-d tuple
        Specifies the kernel size

    number_of_filters : integer
        The number of filters for each encoding layer.

    number_of_residual_blocks : integer
        Number of residual blocks.

    number_of_resnet_blocks : integer
        Number of resnet blocks.  Each block will double the upsampling amount.

    Returns
    -------
    Keras model
        A 2-D Keras model defining the network.

    Example
    -------
    >>> model = create_resnet_super_resolution_model_2d((128, 128, 1))
    >>> model.summary()
    """

    def residual_block_2d(model, number_of_filters, convolution_kernel_size):

        block = Conv2D(filters=number_of_filters,
                       kernel_size=convolution_kernel_size,
                       activation='linear',
                       padding='same')(model)
        block = BatchNormalization()(block)
        block = Activation(activation='relu')(block)
        block = Conv2D(filters=number_of_filters,
                       kernel_size=convolution_kernel_size,
                       activation='linear',
                       padding='same')(block)
        block = BatchNormalization()(block)
        block = Add()([model, block])

        return(block)

    def upscale_block_2d(model, number_of_filters, convolution_kernel_size):

        block = UpSampling2D()(model)
        block = Conv2D(filters=number_of_filters,
                       kernel_size=convolution_kernel_size,
                       activation='relu',
                       padding='same')(block)
        return(block)

    def resnet_block_2d(inputs, number_of_filters, convolution_kernel_size,
                        number_of_residual_blocks):

        outputs = Conv2D(filters=number_of_filters,
                         kernel_size=convolution_kernel_size,
                         activation='relu',
                         padding='same')(inputs)

        residual_blocks = residual_block_2d(outputs, number_of_filters,
                                            convolution_kernel_size)

        for i in range(number_of_residual_blocks):
            residual_blocks = residual_block_2d(residual_blocks, number_of_filters,
                                                convolution_kernel_size)

        outputs = Add()([residual_blocks, outputs])
        outputs = upscale_block_2d(outputs, number_of_filters, convolution_kernel_size)
        return(outputs)

    inputs = Input(shape = input_image_size)

    outputs = resnet_block_2d(inputs, number_of_filters, convolution_kernel_size,
                              number_of_residual_blocks)
    if number_of_resnet_blocks > 1:
        for i in range(1, number_of_resnet_blocks):
             outputs = resnet_block_2d(outputs, number_of_filters,
                                       convolution_kernel_size, number_of_residual_blocks)

    number_of_channels = input_image_size[-1]

    outputs = Conv2D(filters=number_of_channels,
                     kernel_size=convolution_kernel_size,
                     activation='linear',
                     padding='same')(outputs)

    sr_model = Model(inputs=inputs, outputs=outputs)

    return(sr_model)


def create_resnet_super_resolution_model_3d(input_image_size,
                                            convolution_kernel_size=(3, 3, 3),
                                            number_of_filters=64,
                                            number_of_residual_blocks=5,
                                            number_of_resnet_blocks=1
                                           ):
    """
    3-D implementation of the ResNet image super resolution architecture.

    Creates a keras model of the expanded image super resolution deep learning
    framework based on the following python implementation:

            https://github.com/titu1994/Image-Super-Resolution

    Arguments
    ---------
    input_image_size : tuple of length 4
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    convolution_kernel_size : 3-d tuple
        Specifies the kernel size

    number_of_filters : integer
        The number of filters for each encoding layer.

    number_of_residual_blocks : integer
        Number of residual blocks.

    number_of_resnet_blocks : integer
        Number of resnet blocks.  Each block will double the upsampling amount.

    Returns
    -------
    Keras model
        A 3-D Keras model defining the network.

    Example
    -------
    >>> model = create_resnet_super_resolution_model_3d((128, 128, 128, 1))
    >>> model.summary()
    """

    def residual_block_3d(model, number_of_filters, convolution_kernel_size):

        block = Conv3D(filters=number_of_filters,
                       kernel_size=convolution_kernel_size,
                       activation='linear',
                       padding='same')(model)
        block = BatchNormalization()(block)
        block = Activation(activation='relu')(block)
        block = Conv3D(filters=number_of_filters,
                       kernel_size=convolution_kernel_size,
                       activation='linear',
                       padding='same')(block)
        block = BatchNormalization()(block)
        block = Add()([model, block])

        return(block)

    def upscale_block_3d(model, number_of_filters, convolution_kernel_size):

        block = UpSampling3D()(model)
        block = Conv3D(filters=number_of_filters,
                       kernel_size=convolution_kernel_size,
                       activation='relu',
                       padding='same')(block)
        return(block)

    def resnet_block_3d(inputs, number_of_filters, convolution_kernel_size,
                        number_of_residual_blocks):

        outputs = Conv3D(filters=number_of_filters,
                         kernel_size=convolution_kernel_size,
                         activation='relu',
                         padding='same')(inputs)

        residual_blocks = residual_block_3d(outputs, number_of_filters,
                                            convolution_kernel_size)

        for i in range(number_of_residual_blocks):
            residual_blocks = residual_block_3d(residual_blocks, number_of_filters,
                                                convolution_kernel_size)

        outputs = Add()([residual_blocks, outputs])
        outputs = upscale_block_3d(outputs, number_of_filters, convolution_kernel_size)
        return(outputs)

    inputs = Input(shape = input_image_size)

    outputs = resnet_block_3d(inputs, number_of_filters, convolution_kernel_size,
                              number_of_residual_blocks)
    if number_of_resnet_blocks > 1:
        for i in range(1, number_of_resnet_blocks):
             outputs = resnet_block_3d(outputs, number_of_filters,
                                       convolution_kernel_size, number_of_residual_blocks)

    number_of_channels = input_image_size[-1]

    outputs = Conv3D(filters=number_of_channels,
                     kernel_size=convolution_kernel_size,
                     activation='linear',
                     padding='same')(outputs)

    sr_model = Model(inputs=inputs, outputs=outputs)

    return(sr_model)





import numpy as np
import time
import tensorflow as tf
from tensorflow.keras.applications import vgg19
import tensorflow.keras.backend as K

import ants

def neural_style_transfer(content_image,
                          style_images,
                          initial_combination_image=None,
                          number_of_iterations=10,
                          learning_rate=1.0,
                          total_variation_weight=8.5e-5,
                          content_weight=0.025,
                          style_image_weights=1.0,
                          content_layer_names=[
                            'block5_conv2'],
                          style_layer_names="all",
                          content_mask=None,
                          style_masks=None,
                          use_shifted_activations=True,
                          use_chained_inference=True,
                          verbose=False,
                          output_prefix=None):

    """
    The popular neural style transfer described here:

    https://arxiv.org/abs/1508.06576 and https://arxiv.org/abs/1605.04603

    and taken from François Chollet's implementation

    https://keras.io/examples/generative/neural_style_transfer/

    and titu1994's modifications:

    https://github.com/titu1994/Neural-Style-Transfer

    in order to possibly modify and experiment with medical images.

    Arguments
    ---------
    content_image : ANTsImage (1 or 3-component)
        Content (or base) image.

    style_images : ANTsImage or list of ANTsImages
        Style (or reference) image.

    initial_combination_image : ANTsImage (1 or 3-component)
        Starting point for the optimization.  Allows one to start from the
        output from a previous run.  Otherwise, start from the content image.
        Note that the original paper starts with a noise image.

    number_of_iterations : integer
        Number of gradient steps taken during optimization.

    learning_rate : float
        Parameter for Adam optimization.

    total_variation_weight : float
        A penalty on the regularization term to keep the features
        of the output image locally coherent.

    content_weight : float
        Weight of the content layers in the optimization function.

    style_image_weights : float or list of floats
        Weights of the style term in the optimization function for each
        style image.  Can either specify a single scalar to be used for
        all the images or one for each image.  The
        style term computes the sum of the L2 norm between the Gram
        matrices of the different layers (using ImageNet-trained VGG)
        of the style and content images.

    content_layer_names : list of strings
        Names of VGG layers from which to compute the content loss.

    style_layer_names : list of strings
        Names of VGG layers from which to compute the style loss.  If "all",
        the layers used are ['block1_conv1', 'block1_conv2', 'block2_conv1',
        'block2_conv2', 'block3_conv1', 'block3_conv2', 'block3_conv3',
        'block3_conv4', 'block4_conv1', 'block4_conv2', 'block4_conv3',
        'block4_conv4', 'block5_conv1', 'block5_conv2', 'block5_conv3',
        'block5_conv4'].  This is a proposed improvement from
        https://arxiv.org/abs/1605.04603.  In the original implementation, the
        layers used are: ['block1_conv1', 'block2_conv1', 'block3_conv1',
         'block4_conv1', 'block5_conv1'].

    content_mask : ANTsImage
        Specify the region for content consideration.

    style_masks :  ANTsImage or list of ANTsImages
        Specify the region for style consideration.

    use_shifted_activations : boolean
        Use shifted activations in calculating the Gram matrix (improvement
        mentioned in https://arxiv.org/abs/1605.04603).

    use_chained_inference : boolean
        Another proposed improvement from https://arxiv.org/abs/1605.04603.

    verbose : boolean
        Print progress to the screen.

    output_prefix : string
        If specified, outputs a png image to disk at each iteration.

    Returns
    -------
    ANTs 3-component image.

    Example
    -------
    >>> image = neural_style_transfer(content_image, style_image)
    """

    def preprocess_ants_image(image, do_scale_and_center=True):
        array = None
        if image.components == 1:
            array = image.numpy()
            array = np.expand_dims(array, 2)
            array = np.repeat(array, 3, 2)
        elif image.components == 3:
            vector_image = image
            image_channels = ants.split_channels(vector_image)
            array = np.concatenate([np.expand_dims(image_channels[0].numpy(), axis=2),
                                    np.expand_dims(image_channels[1].numpy(), axis=2),
                                    np.expand_dims(image_channels[2].numpy(), axis=2)], axis=2)
        else:
            raise ValueError("Unexpected number of components.")

        if do_scale_and_center == True:
            for i in range(3):
                array[:,:,i] = (array[:,:,i] - array[:,:,i].min()) / (array[:,:,i].max() - array[:,:,i].min())
            array *= 255.0
            # RGB -> BGR
            array = array[:, :, ::-1]
            array[:, :, 0] -= 103.939
            array[:, :, 1] -= 116.779
            array[:, :, 2] -= 123.68

        array = np.expand_dims(array, 0)
        return(array)


    def postprocess_array(array, reference_image):
        array = np.squeeze(array)

        array[:, :, 0] += 103.939
        array[:, :, 1] += 116.779
        array[:, :, 2] += 123.68
        # BGR -> RGB
        array = array[:, :, ::-1]
        array = np.clip(array, 0, 255)

        image = ants.from_numpy(array,
            origin=reference_image.origin, spacing=reference_image.spacing,
            direction=reference_image.direction, has_components=True)
        return(image)


    def gram_matrix(x, shifted_activations=False):
        F = K.batch_flatten(
            K.permute_dimensions(x, (2, 0, 1)))
        if shifted_activations:
            F = F - 1
        gram = K.dot(F, K.transpose(F))
        return(gram)


    def process_mask(mask, shape):
        mask_processed = (tf.image.resize(mask, size=[shape[0], shape[1]],
            method=tf.image.ResizeMethod.NEAREST_NEIGHBOR)).numpy()
        mask_processed_tensor = np.empty(shape)
        for i in range(shape[2]):
            mask_processed_tensor[:, :, i] = mask_processed[:, :, 0]
        return(mask_processed_tensor)

    def style_loss(style_features, combination_features, image_shape, style_mask=None, content_mask=None):

        if content_mask is not None:
            mask_tensor = K.variable(process_mask(content_mask, combination_features.shape))
            combination_features = combination_features * K.stop_gradient(mask_tensor)
            del mask_tensor

        if style_mask is not None:
            mask_tensor = K.variable(process_mask(style_mask, style_features.shape))
            style_features = style_features * K.stop_gradient(mask_tensor)
            if content_mask is not None:
                combination_features = combination_features * K.stop_gradient(mask_tensor)
            del mask_tensor

        style_gram = gram_matrix(style_features, use_shifted_activations)
        content_gram = gram_matrix(combination_features, use_shifted_activations)
        size = image_shape[0] * image_shape[1]
        number_of_channels = 3
        loss = tf.reduce_sum(tf.square(style_gram - content_gram)) / (4.0 * (number_of_channels ** 2) * (size ** 2))
        return(loss)

    def content_loss(content_features, combination_features):
        loss = tf.reduce_sum(tf.square(content_features - combination_features))
        return(loss)

    def total_variation_loss(x):
        shape=x.shape
        a = tf.square(x[:, :(shape[1] - 1), :(shape[2] - 1), :] - x[:, 1:, :(shape[2] - 1), :])
        b = tf.square(x[:, :(shape[1] - 1), :(shape[2] - 1), :] - x[:, :(shape[1] - 1), 1:, :])
        loss = tf.reduce_sum(tf.pow(a + b, 1.25))
        return(loss)

    def compute_total_loss(content_array, style_array_list, combination_tensor,
                           feature_model, content_layer_names, style_layer_names,
                           image_shape, content_mask_tensor=None, style_mask_tensor_list=None):
        number_of_style_images = len(style_array_list)

        input_arrays = list()
        input_arrays.append(content_array)
        for i in range(number_of_style_images):
            input_arrays.append(style_array_list[i])
        input_arrays.append(combination_tensor)
        input_tensor = tf.concat(input_arrays, axis=0)

        features = feature_model(input_tensor)

        total_loss = tf.zeros(shape=())

        # content loss
        for i in range(len(content_layer_names)):
            layer_features = features[content_layer_names[i]]
            content_features = layer_features[0,:, :, :]
            combination_features = layer_features[2, :, :, :]
            total_loss = total_loss + (content_loss(content_features, combination_features) *
               content_weight / len(content_layer_names))

        # style loss
        if use_chained_inference:
            for i in range(len(style_layer_names) - 1):
                layer_features = features[style_layer_names[i]]
                style_features = layer_features[1:(number_of_style_images + 1), :, :, :]
                combination_features = layer_features[number_of_style_images + 1, :, :, :]
                loss = list()
                for j in range(number_of_style_images):
                    if style_mask_tensor_list is None:
                        loss.append(style_loss(style_features[j], combination_features, image_shape,
                                    style_mask=None, content_mask=content_mask_tensor))
                    else:
                        loss.append(style_loss(style_features[j], combination_features, image_shape,
                                    style_mask=style_mask_tensor_list[j], content_mask=content_mask_tensor))

                layer_features = features[style_layer_names[i+1]]
                style_features = layer_features[1:(number_of_style_images + 1), :, :, :]
                combination_features = layer_features[number_of_style_images + 1, :, :, :]
                loss_p1 = list()
                for j in range(number_of_style_images):
                    if style_mask_tensor_list is None:
                        loss_p1.append(style_loss(style_features[j], combination_features, image_shape,
                                       style_mask=None, content_mask=content_mask_tensor))
                    else:
                        loss_p1.append(style_loss(style_features[j], combination_features, image_shape,
                                       style_mask=style_mask_tensor_list[j], content_mask=content_mask_tensor))

                for j in range(number_of_style_images):
                    loss_difference = loss[j] - loss_p1[j]
                    total_loss = total_loss + (style_image_weights[j] * loss_difference /
                        (2 ** (len(style_layer_names) - (i + 1))))

        else:
            for i in range(len(style_layer_names)):
                layer_features = features[style_layer_names[i]]
                style_features = layer_features[1:(number_of_style_images + 1), :, :, :]
                combination_features = layer_features[number_of_style_images + 1, :, :, :]
                for j in range(number_of_style_images):
                    loss = list()
                    if style_mask_tensor_list is None:
                        loss.append(style_loss(style_features[j], combination_features, image_shape,
                                    style_mask=None, content_mask=content_mask_tensor))
                    else:
                        loss.append(style_loss(style_features[j], combination_features, image_shape,
                                    style_mask=style_mask_tensor_list[j], content_mask=content_mask_tensor))

                for j in range(number_of_style_images):
                    total_loss = total_loss + (loss[j] * style_image_weights[j] / len(style_layer_names))


        # total variation loss
        total_loss = total_loss + total_variation_weight * total_variation_loss(combination_tensor)

        return(total_loss)

    def compute_loss_and_gradients(content_array, style_array_list, combination_tensor,
      feature_model, content_layer_names, style_layer_names, image_shape,
      content_mask_tensor=None, style_mask_tensor_list=None):
        with tf.GradientTape() as tape:
            loss = compute_total_loss(content_array, style_array_list, combination_tensor,
                                      feature_model, content_layer_names,
                                      style_layer_names, image_shape, content_mask_tensor,
                                      style_mask_tensor_list)
        gradients = tape.gradient(loss, combination_tensor)
        return loss, gradients

    number_of_style_images = 1
    if isinstance(style_images, list):
        number_of_style_images = len(style_images)

    style_image_list = list()
    if number_of_style_images == 1:
        style_image_list.append(style_images)
    else:
        style_image_list = style_images

    for i in range(number_of_style_images):
        if style_image_list[i].dimension != 2:
            raise ValueError("Input style images must be 2-D.")
        if style_image_list[i].shape != content_image.shape:
            raise ValueError("Input images must have matching dimensions/shapes.")

    number_of_style_masks = 0
    style_mask_tensor_list = None
    if style_masks is not None:
        number_of_style_masks = 1
        if isinstance(style_masks, list):
            number_of_style_masks = len(style_masks)

        style_mask_tensor_list = list()
        if number_of_style_masks == 1:
            style_mask_array = (ants.threshold_image(style_masks, 0, 0, 0, 1)).numpy()
            style_mask_tensor = np.expand_dims(style_mask_array, -1)
            style_mask_tensor_list.append(style_mask_tensor)
        else:
            for i in range(len(style_masks)):
                style_mask_array = (ants.threshold_image(style_masks[i], 0, 0, 0, 1)).numpy()
                style_mask_tensor = np.expand_dims(style_mask_array, -1)
                style_mask_tensor_list.append(style_mask_tensor)

    if number_of_style_masks > 0 and number_of_style_images != number_of_style_masks:
        raise ValueError("The number of style images/masks are not the same.")

    if isinstance(style_image_weights, (int, float)):
        style_image_weights = [style_image_weights] * len(style_image_list)
    else:
        if len(style_image_weights) == 1:
            style_image_weights = style_image_weights * len(style_image_list)
        elif not len(style_image_weights) == len(style_image_list):
            raise ValueError("Length of style weights must be 1 or the number of style images.")

    if content_image.dimension != 2:
        raise ValueError("Input content image must be 2-D.")

    content_mask_tensor = None
    if content_mask is not None:
        content_mask_array = (ants.threshold_image(content_mask, 0, 0, 0, 1)).numpy()
        content_mask_tensor = np.expand_dims(content_mask_array, -1)

    if style_layer_names == "all":
        style_layer_names = ['block1_conv1', 'block1_conv2', 'block2_conv1',
            'block2_conv2', 'block3_conv1', 'block3_conv2', 'block3_conv3',
            'block3_conv4', 'block4_conv1', 'block4_conv2', 'block4_conv3',
            'block4_conv4', 'block5_conv1', 'block5_conv2', 'block5_conv3',
            'block5_conv4']

    model = vgg19.VGG19(weights="imagenet", include_top=False)

    outputs_dictionary = dict([(layer.name, layer.output) for layer in model.layers])
    # shapes_dictionary = dict([(layer.name, layer.output_shape) for layer in model.layers])

    feature_model = tf.keras.Model(inputs=model.inputs, outputs=outputs_dictionary)

    # Preprocess data
    content_array = preprocess_ants_image(content_image)
    style_array_list = list()
    for i in range(number_of_style_images):
        style_array_list.append(preprocess_ants_image(style_image_list[i]))

    image_shape = (content_array.shape[1], content_array.shape[2], 3)

    combination_tensor = None
    if initial_combination_image is None:
        combination_tensor = tf.Variable(np.copy(content_array))
    else:
        initial_combination_tensor = preprocess_ants_image(initial_combination_image, do_scale_and_center=False)
        combination_tensor = tf.Variable(initial_combination_tensor)

    if not image_shape == (combination_tensor.shape[1], combination_tensor.shape[2], 3):
        raise ValueError("Initial combination image size does not match content image.")

    optimizer = tf.optimizers.Adam(learning_rate=learning_rate, beta_1=0.99, epsilon=0.1)

    for i in range(number_of_iterations):
        start_time = time.time()
        loss, gradients = compute_loss_and_gradients(content_array, style_array_list,
                              combination_tensor, feature_model, content_layer_names,
                              style_layer_names, image_shape, content_mask_tensor,
                              style_mask_tensor_list)
        end_time = time.time()
        if verbose == True:
            print("Iteration %d of %d: total loss = %.2f (elapsed time = %ds)" %
              (i, number_of_iterations, loss, end_time - start_time))
        optimizer.apply_gradients([(gradients, combination_tensor)])

        if not output_prefix == None:
            combination_array = combination_tensor.numpy()
            combination_image = postprocess_array(combination_array, content_image)
            combination_rgb = combination_image.vector_to_rgb()
            ants.image_write(combination_rgb, output_prefix + "_iteration%d.png" % (i + 1))

    combination_array = combination_tensor.numpy()
    combination_image = postprocess_array(combination_array, content_image)
    return(combination_image)



import numpy as np
import tensorflow as tf
import ants

def lung_extraction(image,
                    modality="proton",
                    antsxnet_cache_directory=None,
                    verbose=None):

    """
    Perform proton or ct lung extraction using U-net.

    Arguments
    ---------
    image : ANTsImage
        input image

    modality : string
        Modality image type.  Options include "ct" and "proton".

    antsxnet_cache_directory : string
        Destination directory for storing the downloaded template and model weights.
        Since these can be resused, if is None, these data will be downloaded to a
        ~/.keras/ANTsXNet/.

    verbose : boolean
        Print progress to the screen.

    Returns
    -------
    Dictionary of ANTs segmentation and probability images.

    Example
    -------
    >>> output = lung_extraction(lung_image, modality="proton")
    """

    from ..architectures import create_unet_model_3d
    from ..utilities import get_pretrained_network
    from ..utilities import get_antsxnet_data

    if image.dimension != 3:
        raise ValueError( "Image dimension must be 3." )

    if antsxnet_cache_directory == None:
        antsxnet_cache_directory = "ANTsXNet"

    image_mods = [modality]
    channel_size = len(image_mods)

    weights_file_name = None
    unet_model = None

    if modality == "proton":
        weights_file_name = get_pretrained_network("protonLungMri",
            antsxnet_cache_directory=antsxnet_cache_directory)

        classes = ("background", "left_lung", "right_lung")
        number_of_classification_labels = len(classes)

        reorient_template_file_name_path = get_antsxnet_data("protonLungTemplate",
            antsxnet_cache_directory=antsxnet_cache_directory)
        reorient_template = ants.image_read(reorient_template_file_name_path)

        resampled_image_size = reorient_template.shape

        unet_model = create_unet_model_3d((*resampled_image_size, channel_size),
            number_of_outputs = number_of_classification_labels,
            number_of_layers = 4, number_of_filters_at_base_layer = 16, dropout_rate = 0.0,
            convolution_kernel_size = (7, 7, 5), deconvolution_kernel_size = (7, 7, 5))
        unet_model.load_weights(weights_file_name)

        if verbose == True:
            print("Lung extraction:  normalizing image to the template.")

        center_of_mass_template = ants.get_center_of_mass(reorient_template * 0 + 1)
        center_of_mass_image = ants.get_center_of_mass(image * 0 + 1)
        translation = np.asarray(center_of_mass_image) - np.asarray(center_of_mass_template)
        xfrm = ants.create_ants_transform(transform_type="Euler3DTransform",
            center=np.asarray(center_of_mass_template), translation=translation)
        warped_image = ants.apply_ants_transform_to_image(xfrm, image, reorient_template)

        batchX = np.expand_dims(warped_image.numpy(), axis=0)
        batchX = np.expand_dims(batchX, axis=-1)
        batchX = (batchX - batchX.mean()) / batchX.std()

        predicted_data = unet_model.predict(batchX, verbose=0)

        origin = warped_image.origin
        spacing = warped_image.spacing
        direction = warped_image.direction

        probability_images_array = list()
        for i in range(number_of_classification_labels):
            probability_images_array.append(
            ants.from_numpy(np.squeeze(predicted_data[0, :, :, :, i]),
                origin=origin, spacing=spacing, direction=direction))

        if verbose == True:
            print("Lung extraction:  renormalize probability mask to native space.")

        for i in range(number_of_classification_labels):
            probability_images_array[i] = ants.apply_ants_transform_to_image(
                ants.invert_ants_transform(xfrm), probability_images_array[i], image)

        image_matrix = ants.image_list_to_matrix(probability_images_array, image * 0 + 1)
        segmentation_matrix = np.argmax(image_matrix, axis=0)
        segmentation_image = ants.matrix_to_images(
            np.expand_dims(segmentation_matrix, axis=0), image * 0 + 1)[0]

        return_dict = {'segmentation_image' : segmentation_image,
                       'probability_images' : probability_images_array}
        return(return_dict)

    elif modality == "ct":

        ################################
        #
        # Preprocess image
        #
        ################################

        if verbose == True:
            print("Preprocess CT image.")

        def closest_simplified_direction_matrix(direction):
            closest = np.floor(np.abs(direction + 0.5))
            closest[direction < 0] *= -1.0
            return direction

        simplified_direction = closest_simplified_direction_matrix(image.direction)

        reference_image_size = (128, 128, 128)

        ct_preprocessed = ants.resample_image(image, reference_image_size, use_voxels=True, interp_type=0)
        ct_preprocessed[ct_preprocessed < -1000] = -1000
        ct_preprocessed[ct_preprocessed > 400] = 400
        ct_preprocessed.set_direction(simplified_direction)
        ct_preprocessed.set_origin((0, 0, 0))
        ct_preprocessed.set_spacing((1, 1, 1))

        ################################
        #
        # Reorient image
        #
        ################################

        reference_image = ants.make_image(reference_image_size,
                                          voxval=0,
                                          spacing=(1, 1, 1),
                                          origin=(0, 0, 0),
                                          direction=np.identity(3))
        center_of_mass_reference = np.floor(ants.get_center_of_mass(reference_image * 0 + 1))
        center_of_mass_image = np.floor(ants.get_center_of_mass(ct_preprocessed * 0 + 1))
        translation = np.asarray(center_of_mass_image) - np.asarray(center_of_mass_reference)
        xfrm = ants.create_ants_transform(transform_type="Euler3DTransform",
            center=np.asarray(center_of_mass_reference), translation=translation)
        ct_preprocessed = ((ct_preprocessed - ct_preprocessed.min()) /
            (ct_preprocessed.max() - ct_preprocessed.min()))
        ct_preprocessed_warped = ants.apply_ants_transform_to_image(
            xfrm, ct_preprocessed, reference_image, interpolation="nearestneighbor")
        ct_preprocessed_warped = ((ct_preprocessed_warped - ct_preprocessed_warped.min()) /
            (ct_preprocessed_warped.max() - ct_preprocessed_warped.min())) - 0.5

        ################################
        #
        # Build models and load weights
        #
        ################################

        if verbose == True:
            print("Build model and load weights.")

        weights_file_name = get_pretrained_network("lungCtWithPriorsSegmentationWeights", 
            antsxnet_cache_directory=antsxnet_cache_directory)

        classes = ("background", "left lung", "right lung", "airways")
        number_of_classification_labels = len(classes)

        luna16_priors = ants.ndimage_to_list(ants.image_read(get_antsxnet_data("luna16LungPriors")))
        for i in range(len(luna16_priors)):
            luna16_priors[i] = ants.resample_image(luna16_priors[i], reference_image_size, use_voxels=True)
        channel_size = len(luna16_priors) + 1

        unet_model = create_unet_model_3d((*reference_image_size, channel_size),
            number_of_outputs=number_of_classification_labels, mode="classification",
            number_of_layers=4, number_of_filters_at_base_layer=16, dropout_rate=0.0,
            convolution_kernel_size=(3, 3, 3), deconvolution_kernel_size=(2, 2, 2),
            weight_decay=1e-5, additional_options=("attentionGating",))
        unet_model.load_weights(weights_file_name)

        ################################
        #
        # Do prediction and normalize to native space
        #
        ################################

        if verbose == True:
            print("Prediction.")

        batchX = np.zeros((1, *reference_image_size, channel_size))
        batchX[:,:,:,:,0] = ct_preprocessed_warped.numpy()

        for i in range(len(luna16_priors)):
            batchX[:,:,:,:,i+1] = luna16_priors[i].numpy() - 0.5

        predicted_data = unet_model.predict(batchX, verbose=verbose)

        probability_images = list()
        for i in range(number_of_classification_labels):
            if verbose == True:
                print("Reconstructing image", classes[i])
            probability_image = ants.from_numpy(np.squeeze(predicted_data[:,:,:,:,i]), 
                origin=ct_preprocessed_warped.origin, spacing=ct_preprocessed_warped.spacing,
                direction=ct_preprocessed_warped.direction)    
            probability_image = ants.apply_ants_transform_to_image(
                ants.invert_ants_transform(xfrm), probability_image, ct_preprocessed)
            probability_image = ants.resample_image(probability_image,
               resample_params=image.shape, use_voxels=True, interp_type=0)
            probability_image = ants.copy_image_info(image, probability_image)
            probability_images.append(probability_image)

        image_matrix = ants.image_list_to_matrix(probability_images, image * 0 + 1)
        segmentation_matrix = np.argmax(image_matrix, axis=0)
        segmentation_image = ants.matrix_to_images(
            np.expand_dims(segmentation_matrix, axis=0), image * 0 + 1)[0]

        return_dict = {'segmentation_image' : segmentation_image,
                       'probability_images' : probability_images}
        return(return_dict)








import numpy as np
import ants

def deep_flash(t1,
               do_preprocessing=True,
               do_per_hemisphere=True,
               which_hemisphere_models="new",
               antsxnet_cache_directory=None,
               verbose=False
               ):

    """
    Hippocampal/Enthorhinal segmentation using "Deep Flash"

    Perform hippocampal/entorhinal segmentation in T1 images using
    labels from Mike Yassa's lab

    https://faculty.sites.uci.edu/myassa/

    The labeling is as follows:
    Label 0 :  background
    Label 5 :  left aLEC
    Label 6 :  right aLEC
    Label 7 :  left pMEC
    Label 8 :  right pMEC
    Label 9 :  left perirhinal
    Label 10:  right perirhinal
    Label 11:  left parahippocampal
    Label 12:  right parahippocampal
    Label 13:  left DG/CA3
    Label 14:  right DG/CA3
    Label 15:  left CA1
    Label 16:  right CA1
    Label 17:  left subiculum
    Label 18:  right subiculum

    Preprocessing on the training data consisted of:
       * n4 bias correction,
       * denoising,
       * brain extraction, and
       * affine registration to MNI.
    The input T1 should undergo the same steps.  If the input T1 is the raw
    T1, these steps can be performed by the internal preprocessing, i.e. set
    do_preprocessing = True

    Arguments
    ---------
    t1 : ANTsImage
        raw or preprocessed 3-D T1-weighted brain image.

    do_preprocessing : boolean
        See description above.

    do_per_hemisphere : boolean
        If True, do prediction based on separate networks per hemisphere.  Otherwise,
        use the single network trained for both hemispheres.

    antsxnet_cache_directory : string
        Destination directory for storing the downloaded template and model weights.
        Since these can be resused, if is None, these data will be downloaded to a
        ~/.keras/ANTsXNet/.

    verbose : boolean
        Print progress to the screen.

    Returns
    -------
    List consisting of the segmentation image and probability images for
    each label.

    Example
    -------
    >>> image = ants.image_read("t1.nii.gz")
    >>> flash = deep_flash(image)
    """

    from ..architectures import create_unet_model_3d
    from ..utilities import get_pretrained_network
    from ..utilities import get_antsxnet_data
    from ..utilities import categorical_focal_loss
    from ..utilities import preprocess_brain_image
    from ..utilities import crop_image_center
    from ..utilities import pad_or_crop_image_to_size

    if t1.dimension != 3:
        raise ValueError("Image dimension must be 3.")

    if antsxnet_cache_directory == None:
        antsxnet_cache_directory = "ANTsXNet"

    ################################
    #
    # Preprocess images
    #
    ################################

    t1_preprocessed = t1
    if do_preprocessing == True:
        t1_preprocessing = preprocess_brain_image(t1,
            truncate_intensity=(0.01, 0.99),
            brain_extraction_modality="t1",
            template="croppedMni152",
            template_transform_type="antsRegistrationSyNQuickRepro[a]",
            do_bias_correction=True,
            do_denoising=True,
            antsxnet_cache_directory=antsxnet_cache_directory,
            verbose=verbose)
        t1_preprocessed = t1_preprocessing["preprocessed_image"] * t1_preprocessing['brain_mask']

    probability_images = list()
    labels = (0, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18)

    ################################
    #
    # Process left/right in same network
    #
    ################################

    if do_per_hemisphere == False:

        ################################
        #
        # Build model and load weights
        #
        ################################

        template_size = (160, 192, 160)

        unet_model = create_unet_model_3d((*template_size, 1),
            number_of_outputs=len(labels),
            number_of_layers=4, number_of_filters_at_base_layer=8, dropout_rate=0.0,
            convolution_kernel_size=(3, 3, 3), deconvolution_kernel_size=(2, 2, 2),
            weight_decay=1e-5, additional_options=("attentionGating"))

        if verbose == True:
            print("DeepFlash: retrieving model weights.")

        weights_file_name = get_pretrained_network("deepFlash", antsxnet_cache_directory=antsxnet_cache_directory)
        unet_model.load_weights(weights_file_name)

        ################################
        #
        # Do prediction and normalize to native space
        #
        ################################

        if verbose == True:
            print("Prediction.")

        cropped_image = pad_or_crop_image_to_size(t1_preprocessed, template_size)

        batchX = np.expand_dims(cropped_image.numpy(), axis=0)
        batchX = np.expand_dims(batchX, axis=-1)
        batchX = (batchX - batchX.mean()) / batchX.std()

        predicted_data = unet_model.predict(batchX, verbose=verbose)

        origin = cropped_image.origin
        spacing = cropped_image.spacing
        direction = cropped_image.direction

        for i in range(len(labels)):
            probability_image = \
                ants.from_numpy(np.squeeze(predicted_data[0, :, :, :, i]),
                origin=origin, spacing=spacing, direction=direction)
            if i > 0:
                decropped_image = ants.decrop_image(probability_image, t1_preprocessed * 0)
            else:
                decropped_image = ants.decrop_image(probability_image, t1_preprocessed * 0 + 1)

            if do_preprocessing == True:
                probability_images.append(ants.apply_transforms(fixed=t1,
                    moving=decropped_image,
                    transformlist=t1_preprocessing['template_transforms']['invtransforms'],
                    whichtoinvert=[True], interpolator="linear", verbose=verbose))
            else:
                probability_images.append(decropped_image)

    ################################
    #
    # Process left/right in split networks
    #
    ################################

    else:

        ################################
        #
        # Left:  download spatial priors
        #
        ################################

        spatial_priors_left_file_name_path = get_antsxnet_data("priorDeepFlashLeftLabels",
          antsxnet_cache_directory=antsxnet_cache_directory)
        spatial_priors_left = ants.image_read(spatial_priors_left_file_name_path)
        priors_image_left_list = ants.ndimage_to_list(spatial_priors_left)

        ################################
        #
        # Left:  build model and load weights
        #
        ################################

        template_size = (64, 96, 96)
        labels_left = (0, 5, 7, 9, 11, 13, 15, 17)
        channel_size = 1 + len(labels_left)

        number_of_filters = 16
        network_name = ''
        if which_hemisphere_models == "old":
           network_name = "deepFlashLeft16"
        elif which_hemisphere_models == "new":
           network_name = "deepFlashLeft16new"
        else:
            raise ValueError("network_name must be \"old\" or \"new\".")

        unet_model = create_unet_model_3d((*template_size, channel_size),
            number_of_outputs = len(labels_left),
            number_of_layers = 4, number_of_filters_at_base_layer = number_of_filters, dropout_rate = 0.0,
            convolution_kernel_size = (3, 3, 3), deconvolution_kernel_size = (2, 2, 2),
            weight_decay = 1e-5, additional_options=("attentionGating"))

        if verbose == True:
            print("DeepFlash: retrieving model weights (left).")
        weights_file_name = get_pretrained_network(network_name, antsxnet_cache_directory=antsxnet_cache_directory)
        unet_model.load_weights(weights_file_name)

        ################################
        #
        # Left:  do prediction and normalize to native space
        #
        ################################

        if verbose == True:
            print("Prediction (left).")

        cropped_image = ants.crop_indices(t1_preprocessed, (30, 51, 0), (94, 147, 96))
        image_array = cropped_image.numpy()
        image_array = (image_array - image_array.mean()) / image_array.std()

        batchX = np.zeros((1, *template_size, channel_size))
        batchX[0,:,:,:,0] = image_array

        for i in range(len(priors_image_left_list)):
            cropped_prior = ants.crop_indices(priors_image_left_list[i], (30, 51, 0), (94, 147, 96))
            batchX[0,:,:,:,i+1] = cropped_prior.numpy()

        predicted_data = unet_model.predict(batchX, verbose=verbose)

        origin = cropped_image.origin
        spacing = cropped_image.spacing
        direction = cropped_image.direction

        probability_images_left = list()
        for i in range(len(labels_left)):
            probability_image = \
                ants.from_numpy(np.squeeze(predicted_data[0, :, :, :, i]),
                origin=origin, spacing=spacing, direction=direction)
            if i > 0:
                decropped_image = ants.decrop_image(probability_image, t1_preprocessed * 0)
            else:
                decropped_image = ants.decrop_image(probability_image, t1_preprocessed * 0 + 1)

            if do_preprocessing == True:
                probability_images_left.append(ants.apply_transforms(fixed=t1,
                    moving=decropped_image,
                    transformlist=t1_preprocessing['template_transforms']['invtransforms'],
                    whichtoinvert=[True], interpolator="linear", verbose=verbose))
            else:
                probability_images_left.append(decropped_image)

        ################################
        #
        # Right:  download spatial priors
        #
        ################################

        spatial_priors_right_file_name_path = get_antsxnet_data("priorDeepFlashRightLabels",
          antsxnet_cache_directory=antsxnet_cache_directory)
        spatial_priors_right = ants.image_read(spatial_priors_right_file_name_path)
        priors_image_right_list = ants.ndimage_to_list(spatial_priors_right)

        ################################
        #
        # Right:  build model and load weights
        #
        ################################

        template_size = (64, 96, 96)
        labels_right = (0, 6, 8, 10, 12, 14, 16, 18)
        channel_size = 1 + len(labels_right)

        number_of_filters = 16
        network_name = ''
        if which_hemisphere_models == "old":
           network_name = "deepFlashRight16"
        elif which_hemisphere_models == "new":
           network_name = "deepFlashRight16new"
        else:
            raise ValueError("network_name must be \"old\" or \"new\".")

        unet_model = create_unet_model_3d((*template_size, channel_size),
            number_of_outputs = len(labels_right),
            number_of_layers = 4, number_of_filters_at_base_layer = number_of_filters, dropout_rate = 0.0,
            convolution_kernel_size = (3, 3, 3), deconvolution_kernel_size = (2, 2, 2),
            weight_decay = 1e-5, additional_options=("attentionGating"))

        weights_file_name = get_pretrained_network(network_name, antsxnet_cache_directory=antsxnet_cache_directory)
        unet_model.load_weights(weights_file_name)

        ################################
        #
        # Right:  do prediction and normalize to native space
        #
        ################################

        if verbose == True:
            print("Prediction (right).")

        cropped_image = ants.crop_indices(t1_preprocessed, (88, 51, 0), (152, 147, 96))
        image_array = cropped_image.numpy()
        image_array = (image_array - image_array.mean()) / image_array.std()

        batchX = np.zeros((1, *template_size, channel_size))
        batchX[0,:,:,:,0] = image_array

        for i in range(len(priors_image_right_list)):
            cropped_prior = ants.crop_indices(priors_image_right_list[i], (88, 51, 0), (152, 147, 96))
            batchX[0,:,:,:,i+1] = cropped_prior.numpy()

        predicted_data = unet_model.predict(batchX, verbose=verbose)

        origin = cropped_image.origin
        spacing = cropped_image.spacing
        direction = cropped_image.direction

        probability_images_right = list()
        for i in range(len(labels_right)):
            probability_image = \
                ants.from_numpy(np.squeeze(predicted_data[0, :, :, :, i]),
                origin=origin, spacing=spacing, direction=direction)
            if i > 0:
                decropped_image = ants.decrop_image(probability_image, t1_preprocessed * 0)
            else:
                decropped_image = ants.decrop_image(probability_image, t1_preprocessed * 0 + 1)

            if do_preprocessing == True:
                probability_images_right.append(ants.apply_transforms(fixed=t1,
                    moving=decropped_image,
                    transformlist=t1_preprocessing['template_transforms']['invtransforms'],
                    whichtoinvert=[True], interpolator="linear", verbose=verbose))
            else:
                probability_images_right.append(decropped_image)

        ################################
        #
        # Combine priors
        #
        ################################

        probability_background_image = ants.image_clone(t1) * 0
        for i in range(1, len(probability_images_left)):
            probability_background_image += probability_images_left[i]
        for i in range(1, len(probability_images_right)):
            probability_background_image += probability_images_right[i]

        probability_images.append(probability_background_image * -1 + 1)
        for i in range(1, len(probability_images_left)):
            probability_images.append(probability_images_left[i])
            probability_images.append(probability_images_right[i])

    ################################
    #
    # Convert probability images to segmentation
    #
    ################################

    # image_matrix = ants.image_list_to_matrix(probability_images, t1 * 0 + 1)
    # segmentation_matrix = np.argmax(image_matrix, axis=0)
    # segmentation_image = ants.matrix_to_images(
    #     np.expand_dims(segmentation_matrix, axis=0), t1 * 0 + 1)[0]

    image_matrix = ants.image_list_to_matrix(probability_images[1:(len(probability_images))], t1 * 0 + 1)
    background_foreground_matrix = np.stack([ants.image_list_to_matrix([probability_images[0]], t1 * 0 + 1),
                                            np.expand_dims(np.sum(image_matrix, axis=0), axis=0)])
    foreground_matrix = np.argmax(background_foreground_matrix, axis=0)
    segmentation_matrix = (np.argmax(image_matrix, axis=0) + 1) * foreground_matrix
    segmentation_image = ants.matrix_to_images(
        np.expand_dims(segmentation_matrix, axis=0), t1 * 0 + 1)[0]

    relabeled_image = ants.image_clone(segmentation_image)
    for i in range(len(labels)):
        relabeled_image[segmentation_image==i] = labels[i]

    return_dict = {'segmentation_image' : relabeled_image,
                   'probability_images' : probability_images}
    return(return_dict)



import numpy as np
import ants

def deep_atropos(t1,
                 do_preprocessing=True,
                 use_spatial_priors=1,
                 antsxnet_cache_directory=None,
                 verbose=False):

    """
    Six-tissue segmentation.

    Perform Atropos-style six tissue segmentation using deep learning.

    The labeling is as follows:
    Label 0 :  background
    Label 1 :  CSF
    Label 2 :  gray matter
    Label 3 :  white matter
    Label 4 :  deep gray matter
    Label 5 :  brain stem
    Label 6 :  cerebellum

    Preprocessing on the training data consisted of:
       * n4 bias correction,
       * denoising,
       * brain extraction, and
       * affine registration to MNI.
    The input T1 should undergo the same steps.  If the input T1 is the raw
    T1, these steps can be performed by the internal preprocessing, i.e. set
    do_preprocessing = True

    Arguments
    ---------
    t1 : ANTsImage
        raw or preprocessed 3-D T1-weighted brain image.

    do_preprocessing : boolean
        See description above.

    use_spatial_priors : integer
        Use MNI spatial tissue priors (0 or 1).  Currently, only '0' (no priors) and '1'
        (cerebellar prior only) are the only two options.  Default is 1.

    antsxnet_cache_directory : string
        Destination directory for storing the downloaded template and model weights.
        Since these can be resused, if is None, these data will be downloaded to a
        ~/.keras/ANTsXNet/.

    verbose : boolean
        Print progress to the screen.

    Returns
    -------
    List consisting of the segmentation image and probability images for
    each label.

    Example
    -------
    >>> image = ants.image_read("t1.nii.gz")
    >>> flash = deep_atropos(image)
    """

    from ..architectures import create_unet_model_3d
    from ..utilities import get_pretrained_network
    from ..utilities import get_antsxnet_data
    from ..utilities import preprocess_brain_image
    from ..utilities import extract_image_patches
    from ..utilities import reconstruct_image_from_patches

    if t1.dimension != 3:
        raise ValueError( "Image dimension must be 3." )

    if antsxnet_cache_directory == None:
        antsxnet_cache_directory = "ANTsXNet"

    ################################
    #
    # Preprocess images
    #
    ################################

    t1_preprocessed = t1
    if do_preprocessing == True:
        t1_preprocessing = preprocess_brain_image(t1,
            truncate_intensity=(0.01, 0.99),
            brain_extraction_modality="t1",
            template="croppedMni152",
            template_transform_type="antsRegistrationSyNQuickRepro[a]",
            do_bias_correction=True,
            do_denoising=True,
            antsxnet_cache_directory=antsxnet_cache_directory,
            verbose=verbose)
        t1_preprocessed = t1_preprocessing["preprocessed_image"] * t1_preprocessing['brain_mask']

    ################################
    #
    # Build model and load weights
    #
    ################################

    patch_size = (112, 112, 112)
    stride_length = (t1_preprocessed.shape[0] - patch_size[0],
                     t1_preprocessed.shape[1] - patch_size[1],
                     t1_preprocessed.shape[2] - patch_size[2])

    classes = ("background", "csf", "gray matter", "white matter",
               "deep gray matter", "brain stem", "cerebellum")

    mni_priors = None
    channel_size = 1
    if use_spatial_priors != 0:
        mni_priors = ants.ndimage_to_list(ants.image_read(get_antsxnet_data("croppedMni152Priors")))
        for i in range(len(mni_priors)):
            mni_priors[i] = ants.copy_image_info(t1_preprocessed, mni_priors[i])
        channel_size = 2

    unet_model = create_unet_model_3d((*patch_size, channel_size),
        number_of_outputs=len(classes), mode = "classification",
        number_of_layers=4, number_of_filters_at_base_layer=16, dropout_rate=0.0,
        convolution_kernel_size=(3, 3, 3), deconvolution_kernel_size=(2, 2, 2),
        weight_decay=1e-5, additional_options=("attentionGating"))

    if verbose == True:
        print("DeepAtropos:  retrieving model weights.")

    weights_file_name = ''
    if use_spatial_priors == 0:
        weights_file_name = get_pretrained_network("sixTissueOctantBrainSegmentation", antsxnet_cache_directory=antsxnet_cache_directory)
    elif use_spatial_priors == 1:
        weights_file_name = get_pretrained_network("sixTissueOctantBrainSegmentationWithPriors1", antsxnet_cache_directory=antsxnet_cache_directory)
    else:
        raise ValueError("use_spatial_priors must be a 0 or 1")
    unet_model.load_weights(weights_file_name)

    ################################
    #
    # Do prediction and normalize to native space
    #
    ################################

    if verbose == True:
        print("Prediction.")

    t1_preprocessed = (t1_preprocessed - t1_preprocessed.mean()) / t1_preprocessed.std()
    image_patches = extract_image_patches(t1_preprocessed, patch_size=patch_size,
                                          max_number_of_patches="all", stride_length=stride_length,
                                          return_as_array=True)
    batchX = np.zeros((*image_patches.shape, channel_size))
    batchX[:,:,:,:,0] = image_patches
    if channel_size > 1:
        prior_patches = extract_image_patches(mni_priors[6], patch_size=patch_size,
                            max_number_of_patches="all", stride_length=stride_length,
                            return_as_array=True)
        batchX[:,:,:,:,1] = prior_patches

    predicted_data = unet_model.predict(batchX, verbose=verbose)

    probability_images = list()
    for i in range(len(classes)):
        if verbose == True:
            print("Reconstructing image", classes[i])
        reconstructed_image = reconstruct_image_from_patches(predicted_data[:,:,:,:,i],
            domain_image=t1_preprocessed, stride_length=stride_length)

        if do_preprocessing == True:
            probability_images.append(ants.apply_transforms(fixed=t1,
                moving=reconstructed_image,
                transformlist=t1_preprocessing['template_transforms']['invtransforms'],
                whichtoinvert=[True], interpolator="linear", verbose=verbose))
        else:
            probability_images.append(reconstructed_image)

    image_matrix = ants.image_list_to_matrix(probability_images, t1 * 0 + 1)
    segmentation_matrix = np.argmax(image_matrix, axis=0)
    segmentation_image = ants.matrix_to_images(
        np.expand_dims(segmentation_matrix, axis=0), t1 * 0 + 1)[0]

    return_dict = {'segmentation_image' : segmentation_image,
                   'probability_images' : probability_images}
    return(return_dict)



import tensorflow as tf

import tensorflow.keras.backend as K

from tensorflow.keras.models import Model, Sequential
from tensorflow.keras.layers import Layer
from tensorflow.keras.layers import (Input, Concatenate, Dense, Activation,
                                     BatchNormalization, Reshape, Dropout,
                                     Flatten, LeakyReLU, Conv2D, Conv3D,
                                     UpSampling2D, UpSampling3D)
from tensorflow.keras.optimizers import Adam

from ..utilities import InstanceNormalization

import numpy as np
import os

import matplotlib.pyplot as plot

import ants

class CycleGanModel(object):
    """
    Cycle GAN model

    Cycle generative adverserial network from the paper:

      https://arxiv.org/pdf/1703.10593

    and ported from the Keras (python) implementation:

      https://github.com/eriklindernoren/Keras-GAN/blob/master/cyclegan/cyclegan.py

    Arguments
    ---------
    input_image_size : tuple
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    latent_dimension : integer

    Returns
    -------
    Keras model
        A Keras model defining the network.
    """

    def __init__(self, input_image_size, lambda_cycle_loss_weight=10.0,
                 lambda_identity_loss_weight=1.0,
                 number_of_filters_at_base_layer=(32, 64)):
        super(CycleGanModel, self).__init__()

        self.input_image_size = input_image_size
        self.number_of_channels = self.input_image_size[-1]
        self.discriminator_patch_size = None
        self.lambda_cycle_loss_weight = lambda_cycle_loss_weight
        self.lambda_identity_loss_weight = lambda_identity_loss_weight

        self.number_of_filters_at_base_layer = number_of_filters_at_base_layer

        self.dimensionality = None
        if len(self.input_image_size) == 3:
            self.dimensionality = 2
        elif len(self.input_image_size) == 4:
            self.dimensionality = 3
        else:
            raise ValueError("Incorrect size for input_image_size.")

        optimizer = Adam(lr=0.0002, beta_1=0.5)

        # Build discriminators for domains A and B

        self.discriminatorA = self.build_discriminator()
        self.discriminatorA.compile(loss='mse',
                                   optimizer=optimizer, metrics=['acc'])
        self.discriminatorA.trainable = False

        self.discriminatorB = self.build_discriminator()
        self.discriminatorB.compile(loss='mse',
                                   optimizer=optimizer, metrics=['acc'])
        self.discriminatorB.trainable = False

        # Build u-net like generators

        self.generatorAtoB = self.build_generator()
        self.generatorBtoA = self.build_generator()

        imageA = Input(shape=input_image_size)
        imageB = Input(shape=input_image_size)

        fake_imageA = self.generatorBtoA(imageB)
        fake_imageB = self.generatorAtoB(imageA)

        reconstructed_imageA = self.generatorBtoA(fake_imageB)
        reconstructed_imageB = self.generatorAtoB(fake_imageA)

        identity_imageA = self.generatorBtoA(imageA)
        identity_imageB = self.generatorAtoB(imageB)

        # Check images

        validityA = self.discriminatorA(fake_imageA)
        validityB = self.discriminatorB(fake_imageB)

        # Combined models

        self.combined_model = Model(inputs=[imageA, imageB],
                                    outputs=[validityA, validityB,
                                             reconstructed_imageA, reconstructed_imageB,
                                             identity_imageA, identity_imageB])
        self.combined_model.compile(loss=['mse', 'mse', 'mae', 'mae', 'mae', 'mae'],
                                    loss_weights=[1.0, 1.0,
                                                  self.lambda_cycle_loss_weight, self.lambda_cycle_loss_weight,
                                                  self.lambda_identity_loss_weight, self.lambda_identity_loss_weight],
                                    optimizer=optimizer)

    def build_generator(self):

        def build_encoding_layer(input, number_of_filters, kernel_size=4):
            encoder = input
            if self.dimensionality == 2:
                encoder = Conv2D(filters=number_of_filters,
                                 kernel_size=kernel_size,
                                 strides=2,
                                 padding='same')(encoder)
            else:
                encoder = Conv3D(filters=number_of_filters,
                                 kernel_size=kernel_size,
                                 strides=2,
                                 padding='same')(encoder)
            encoder = LeakyReLU(alpha=0.2)(encoder)
            encoder = InstanceNormalization()(encoder)
            return(encoder)

        def build_decoding_layer(input, skip_input, number_of_filters,
                                 kernel_size=4, dropout_rate=0.0):
            decoder = input
            if self.dimensionality == 2:
                decoder = UpSampling2D(size=2)(decoder)
                decoder = Conv2D(filters=number_of_filters,
                                 kernel_size=kernel_size,
                                 strides=1,
                                 padding='same',
                                 activation='relu')(decoder)
            else:
                decoder = UpSampling3D(size=2)(decoder)
                decoder = Conv3D(filters=number_of_filters,
                                 kernel_size=kernel_size,
                                 strides=1,
                                 padding='same',
                                 activation='relu')(decoder)
            if dropout_rate > 0.0:
                decoder = Dropout(dropout_rate=dropout_rate)(decoder)
            decoder = LeakyReLU(alpha=0.2)(decoder)
            decoder = Concatenate()([decoder, skip_input])
            return(decoder)

        input = Input(shape=self.input_image_size)

        encoding_layers = list()
        encoding_layers.append(build_encoding_layer(input,
                                                    int(self.number_of_filters_at_base_layer[0])))
        encoding_layers.append(build_encoding_layer(encoding_layers[0],
                                                    int(self.number_of_filters_at_base_layer[0] * 2)))
        encoding_layers.append(build_encoding_layer(encoding_layers[1],
                                                    int(self.number_of_filters_at_base_layer[0] * 4)))
        encoding_layers.append(build_encoding_layer(encoding_layers[2],
                                                    int(self.number_of_filters_at_base_layer[0] * 8)))

        decoding_layers = list()
        decoding_layers.append(build_decoding_layer(encoding_layers[3], encoding_layers[2],
                                                    int(self.number_of_filters_at_base_layer[0] * 4)))
        decoding_layers.append(build_decoding_layer(decoding_layers[0], encoding_layers[1],
                                                    int(self.number_of_filters_at_base_layer[0] * 2)))
        decoding_layers.append(build_decoding_layer(decoding_layers[1], encoding_layers[0],
                                                    int(self.number_of_filters_at_base_layer[0])))

        if self.dimensionality == 2:
            decoding_layers.append(UpSampling2D(size=2)(decoding_layers[-1]))
            decoding_layers[-1] = Conv2D(filters=self.number_of_channels,
                                         kernel_size=4,
                                         strides=1,
                                         padding='same',
                                         activation='tanh')(decoding_layers[-1])
        else:
            decoding_layers.append(UpSampling3D(size=2)(decoding_layers[-1]))
            decoding_layers[-1] = Conv2D(filters=self.number_of_channels,
                                         kernel_size=4,
                                         strides=1,
                                         padding='same',
                                         activation='tanh')(decoding_layers[-1])

        generator = Model(inputs=input, outputs=decoding_layers[-1])
        return(generator)

    def build_discriminator(self):

        def build_layer(input, number_of_filters, kernel_size=4, normalization=True):
            layer = input
            if self.dimensionality == 2:
                layer = Conv2D(filters=number_of_filters,
                                 kernel_size=kernel_size,
                                 strides=2,
                                 padding='same')(layer)
            else:
                layer = Conv3D(filters=number_of_filters,
                                 kernel_size=kernel_size,
                                 strides=2,
                                 padding='same')(layer)
            layer = LeakyReLU(alpha=0.2)(layer)
            if normalization == True:
                layer = InstanceNormalization()(layer)
            return(layer)

        input = Input(shape=self.input_image_size)

        layers = list()
        layers.append(build_layer(input,
                                  int(self.number_of_filters_at_base_layer[1])))
        layers.append(build_layer(layers[0],
                                  int(self.number_of_filters_at_base_layer[1] * 2)))
        layers.append(build_layer(layers[1],
                                  int(self.number_of_filters_at_base_layer[1] * 4)))
        layers.append(build_layer(layers[2],
                                  int(self.number_of_filters_at_base_layer[1] * 8)))

        validity = None
        if self.dimensionality == 2:
            validity = Conv2D(filters=1,
                              kernel_size=4,
                              strides=1,
                              padding='same')(layers[3])
        else:
            validity = Conv3D(filters=1,
                              kernel_size=4,
                              strides=1,
                              padding='same')(layers[3])

        if self.discriminator_patch_size is None:
            self.discriminator_patch_size = K.int_shape(validity)[1:]

        discriminator = Model(inputs=input, outputs=validity)
        return(discriminator)

    def train(self, X_trainA, X_trainB, number_of_epochs, batch_size=128,
              sample_interval=None, sample_file_prefix='sample'):

        valid = np.ones((batch_size, *self.discriminator_patch_size))
        fake = np.zeros((batch_size, *self.discriminator_patch_size))

        for epoch in range(number_of_epochs):

            indicesA = np.random.randint(0, X_trainA.shape[0] - 1, batch_size)
            imagesA = X_trainA[indicesA]

            indicesB = np.random.randint(0, X_trainB.shape[0] - 1, batch_size)
            imagesB = X_trainB[indicesB]

            # train discriminator

            fake_imagesA = self.generatorAtoB.predict(imagesA)
            fake_imagesB = self.generatorBtoA.predict(imagesB)

            dA_loss_real = self.discriminatorA.train_on_batch(imagesA, valid)
            dA_loss_fake = self.discriminatorA.train_on_batch(fake_imagesA, fake)

            dB_loss_real = self.discriminatorB.train_on_batch(imagesB, valid)
            dB_loss_fake = self.discriminatorB.train_on_batch(fake_imagesB, fake)

            d_loss = list()
            for i in range(len(dA_loss_real)):
                d_loss.append(0.25 * (dA_loss_real[i] + dA_loss_fake[i] +
                                      dB_loss_real[i] + dB_loss_fake[i]))

            # train generator

            g_loss = self.combined_model.train_on_batch([imagesA, imagesB],
              [valid, valid, imagesA, imagesB, imagesA, imagesB])

            print("Epoch ", epoch, ": [Discriminator loss: ", d_loss[0],
                  " acc: ", d_loss[1], "] ", "[Generator loss: ", g_loss[0],
                  ", ", np.mean(g_loss[1:3]), ", ", np.mean(g_loss[3:5]), ", ",
                  np.mean(g_loss[5:6]), "]")

            if self.dimensionality == 2:
                if sample_interval != None:
                    if epoch % sample_interval == 0:

                        # Do a 2x3 grid
                        #
                        # imageA  |  translated( imageA ) | reconstructed( imageA )
                        # imageB  |  translated( imageB ) | reconstructed( imageB )

                        indexA = np.random.randint(0, X_trainA.shape[0] - 1, 1)
                        indexB = np.random.randint(0, X_trainB.shape[0] - 1, 1)

                        imageA = X_trainA[indexA,:,:,:]
                        imageB = X_trainB[indexB,:,:,:]

                        X = list()
                        X.append(imageA)
                        X.append(self.generatorAtoB.predict(X[0]))
                        X.append(self.generatorBtoA.predict(X[1]))

                        X.append(imageB)
                        X.append(self.generatorAtoB.predict(X[3]))
                        X.append(self.generatorBtoA.predict(X[4]))

                        plot_images = np.concatenate(X)
                        plot_images = 0.5 * plot_images + 0.5

                        titles = ['Original', 'Translated', 'Reconstructed']
                        figure, axes = plot.subplots(2, 3)

                        count = 0
                        for i in range(2):
                            for j in range(3):
                                axes[i, j].imshow(plot_images[count])
                                axes[i, j].set_title(titles[j])
                                axes[i, j].axis('off')
                                count += 1

                        image_file_name = sample_file_prefix + "_iteration" + str(epoch) + ".jpg"
                        dir_name = os.path.dirname(sample_file_prefix)
                        if not os.path.exists(dir_name):
                            os.mkdir(dir_name)
                        figure.savefig(image_file_name)
                        plot.close()





from tensorflow.keras.models import Model
from tensorflow.keras.layers import (Input, Add, Subtract,
                          PReLU, Concatenate,
                          Conv2D, Conv2DTranspose,
                          Conv3D, Conv3DTranspose)

def create_deep_back_projection_network_model_2d(input_image_size,
                                                 number_of_outputs=1,
                                                 number_of_base_filters=64,
                                                 number_of_feature_filters=256,
                                                 number_of_back_projection_stages=7,
                                                 convolution_kernel_size=(12, 12),
                                                 strides=(8, 8),
                                                 last_convolution=(3, 3),
                                                 number_of_loss_functions=1
                                                ):
    """
    2-D implementation of the deep back-projection network.

    Creates a keras model of the deep back-project network for image super
    resolution.  More information is provided at the authors' website:

            https://www.toyota-ti.ac.jp/Lab/Denshi/iim/members/muhammad.haris/projects/DBPN.html

    with the paper available here:

            https://arxiv.org/abs/1803.02735

    This particular implementation was influenced by the following keras (python)
    implementation:

            https://github.com/rajatkb/DBPN-Keras

    with help from the original author's Caffe and Pytorch implementations:

            https://github.com/alterzero/DBPN-caffe
            https://github.com/alterzero/DBPN-Pytorch

    Arguments
    ---------
    input_image_size : tuple of length 3
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    number_of_outputs : integer
        Number of outputs (e.g., 3 for RGB images).

    number_of_feature_filters : integer
        Number of feature filters.

    number_of_base_filters : integer
        Number of base filters.

    number_of_back_projection_stages : integer
        Number of up-down-projection stages.
        This number includes the final up block.

    convolution_kernel_size : tuple of length 2
        Kernel size for certain convolutional layers.  The strides are dependent on
        the scale factor discussed in original paper.  Factors used in the original
        implementation are as follows:
        2x --> convolution_kernel_size=(6, 6),
        4x --> convolution_kernel_size=(8, 8),
        8x --> convolution_kernel_size=(12, 12).  We default to 8x parameters.

    strides : tuple of length 2
        Strides for certain convolutional layers.  This and the
        convolution_kernel_size are dependent on the scale factor discussed in
        original paper.  Factors used in the original implementation are as follows:
        2x --> strides = (2, 2),
        4x --> strides = (4, 4),
        8x --> strides = (8, 8). We default to 8x parameters.

    last_convolution: tuple of length 2
        The kernel size for the last convolutional layer.

    number_of_loss_functions :  integer
        The number of data targets, e.g. 2 for 2 targets

    Returns
    -------
    Keras model
        A 2-D Keras model defining the network.

    Example
    -------
    >>> model = create_deep_back_projection_network_model_2d((128, 128, 1))
    >>> model.summary()
    """

    def up_block_2d(L, number_of_filters=64, kernel_size=(12, 12), strides=(8, 8),
                    include_dense_convolution_layer=True):
        if include_dense_convolution_layer == True:
            L = Conv2D(filters = number_of_filters,
                       use_bias=True,
                       kernel_size=(1, 1),
                       strides=(1, 1),
                       padding='same')(L)
            L = PReLU(alpha_initializer='zero',
                      shared_axes=[1, 2])(L)

        # Scale up
        H0 = Conv2DTranspose(filters=number_of_filters,
                             kernel_size=kernel_size,
                             strides=strides,
                             kernel_initializer='glorot_uniform',
                             padding='same')(L)
        H0 = PReLU(alpha_initializer='zero',
                   shared_axes=[1, 2])(H0)

        # Scale down
        L0 = Conv2D(filters=number_of_filters,
                    kernel_size=kernel_size,
                    strides=strides,
                    kernel_initializer='glorot_uniform',
                    padding='same')(H0)
        L0 = PReLU(alpha_initializer='zero',
                   shared_axes=[1, 2])(L0)

        # Residual
        E = Subtract()([L0, L])

        # Scale residual up
        H1 = Conv2DTranspose(filters=number_of_filters,
                             kernel_size=kernel_size,
                             strides=strides,
                             kernel_initializer='glorot_uniform',
                             padding='same')(E)
        H1 = PReLU(alpha_initializer='zero',
                   shared_axes=[1, 2])(H1)

        # Output feature map
        up_block = Add()([H0, H1])

        return(up_block)


    def down_block_2d(H, number_of_filters=64, kernel_size=(12, 12), strides=(8, 8),
                    include_dense_convolution_layer=True):
        if include_dense_convolution_layer == True:
            H = Conv2D(filters = number_of_filters,
                       use_bias=True,
                       kernel_size=(1, 1),
                       strides=(1, 1),
                       padding='same')(H)
            H = PReLU(alpha_initializer='zero',
                      shared_axes=[1, 2])(H)

        # Scale down
        L0 = Conv2D(filters=number_of_filters,
                    kernel_size=kernel_size,
                    strides=strides,
                    kernel_initializer='glorot_uniform',
                    padding='same')(H)
        L0 = PReLU(alpha_initializer='zero',
                   shared_axes=[1, 2])(L0)

        # Scale up
        H0 = Conv2DTranspose(filters=number_of_filters,
                             kernel_size=kernel_size,
                             strides=strides,
                             kernel_initializer='glorot_uniform',
                             padding='same')(L0)
        H0 = PReLU(alpha_initializer='zero',
                   shared_axes=[1, 2])(H0)

        # Residual
        E = Subtract()([H0, H])

        # Scale residual down
        L1 = Conv2D(filters=number_of_filters,
                    kernel_size=kernel_size,
                    strides=strides,
                    kernel_initializer='glorot_uniform',
                    padding='same')(E)
        L1 = PReLU(alpha_initializer='zero',
                   shared_axes=[1, 2])(L1)

        # Output feature map
        down_block = Add()([L0, L1])

        return(down_block)

    inputs = Input(shape=input_image_size)

    # Initial feature extraction
    model = Conv2D(filters=number_of_feature_filters,
                   kernel_size=(3, 3),
                   strides=(1, 1),
                   padding='same',
                   kernel_initializer='glorot_uniform')(inputs)
    model = PReLU(alpha_initializer='zero',
                  shared_axes=[1, 2])(model)

    # Feature smashing
    model = Conv2D(filters=number_of_base_filters,
                   kernel_size=(1, 1),
                   strides=(1, 1),
                   padding='same',
                   kernel_initializer='glorot_uniform')(model)
    model = PReLU(alpha_initializer='zero',
                  shared_axes=[1, 2])(model)

    # Back projection
    up_projection_blocks = []
    down_projection_blocks = []

    model = up_block_2d(model, number_of_filters=number_of_base_filters,
      kernel_size=convolution_kernel_size, strides=strides)
    up_projection_blocks.append(model)

    for i in range(number_of_back_projection_stages):
        if i == 0:
            model = down_block_2d(model, number_of_filters=number_of_base_filters,
              kernel_size=convolution_kernel_size, strides=strides)
            down_projection_blocks.append(model)

            model = up_block_2d(model, number_of_filters=number_of_base_filters,
              kernel_size=convolution_kernel_size, strides=strides)
            up_projection_blocks.append(model)

            model = Concatenate()(up_projection_blocks)
        else:
            model = down_block_2d(model, number_of_filters=number_of_base_filters,
              kernel_size=convolution_kernel_size, strides=strides,
              include_dense_convolution_layer=True)
            down_projection_blocks.append(model)
            model = Concatenate()(down_projection_blocks)

            model = up_block_2d(model, number_of_filters=number_of_base_filters,
              kernel_size=convolution_kernel_size, strides=strides,
              include_dense_convolution_layer=True)
            up_projection_blocks.append(model)

            model = Concatenate()(up_projection_blocks)

    # Final convolution layer
    outputs = Conv2D(filters=number_of_outputs,
                     kernel_size=last_convolution,
                     strides=(1, 1),
                     padding = 'same',
                     kernel_initializer = "glorot_uniform")(model)

    if number_of_loss_functions == 1:
        deep_back_projection_network_model = Model(inputs=inputs, outputs=outputs)
    else:
        outputList=[]
        for k in range(number_of_loss_functions):
            outputList.append(outputs)
        deep_back_projection_network_model = Model(inputs=inputs, outputs=outputList)

    return(deep_back_projection_network_model)


def create_deep_back_projection_network_model_3d(input_image_size,
                                                 number_of_outputs=1,
                                                 number_of_base_filters=64,
                                                 number_of_feature_filters=256,
                                                 number_of_back_projection_stages=7,
                                                 convolution_kernel_size=(12, 12, 12),
                                                 strides=(8, 8, 8),
                                                 last_convolution=(3, 3, 3),
                                                 number_of_loss_functions=1
                                                ):
    """
    3-D implementation of the deep back-projection network.

    Creates a keras model of the deep back-project network for image super
    resolution.  More information is provided at the authors' website:

            https://www.toyota-ti.ac.jp/Lab/Denshi/iim/members/muhammad.haris/projects/DBPN.html

    with the paper available here:

            https://arxiv.org/abs/1803.02735

    This particular implementation was influenced by the following keras (python)
    implementation:

            https://github.com/rajatkb/DBPN-Keras

    with help from the original author's Caffe and Pytorch implementations:

            https://github.com/alterzero/DBPN-caffe
            https://github.com/alterzero/DBPN-Pytorch

    Arguments
    ---------
    input_image_size : tuple of length 4
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    number_of_outputs : integer
        Number of outputs (e.g., 3 for RGB images).

    number_of_feature_filters : integer
        Number of feature filters.

    number_of_base_filters : integer
        Number of base filters.

    number_of_back_projection_stages : integer
        Number of up-down-projection stages.
        This number includes the final up block.

    convolution_kernel_size : tuple of length 3
        Kernel size for certain convolutional layers.  The strides are dependent on
        the scale factor discussed in original paper.  Factors used in the original
        implementation are as follows:
        2x --> convolution_kernel_size=(6, 6, 6),
        4x --> convolution_kernel_size=(8, 8, 8),
        8x --> convolution_kernel_size=(12, 12, 12).  We default to 8x parameters.

    strides : tuple of length 3
        Strides for certain convolutional layers.  This and the
        convolution_kernel_size are dependent on the scale factor discussed in
        original paper.  Factors used in the original implementation are as follows:
        2x --> strides = (2, 2, 2),
        4x --> strides = (4, 4, 4),
        8x --> strides = (8, 8, 8). We default to 8x parameters.

    last_convolution: tuple of length 3
        The kernel size for the last convolutional layer.

    number_of_loss_functions :  integer
        The number of data targets, e.g. 2 for 2 targets

    Returns
    -------
    Keras model
        A 3-D Keras model defining the network.

    Example
    -------
    >>> model = create_deep_back_projection_network_model_3d((128, 128, 128, 1))
    >>> model.summary()
    """

    def up_block_3d(L, number_of_filters=64, kernel_size=(12, 12, 12), strides=(8, 8, 8),
                    include_dense_convolution_layer=True):
        if include_dense_convolution_layer == True:
            L = Conv3D(filters = number_of_filters,
                       use_bias=True,
                       kernel_size=(1, 1, 1),
                       strides=(1, 1, 1),
                       padding='same')(L)
            L = PReLU(alpha_initializer='zero',
                      shared_axes=[1, 2, 3])(L)

        # Scale up
        H0 = Conv3DTranspose(filters=number_of_filters,
                             kernel_size=kernel_size,
                             strides=strides,
                             kernel_initializer='glorot_uniform',
                             padding='same')(L)
        H0 = PReLU(alpha_initializer='zero',
                   shared_axes=[1, 2, 3])(H0)

        # Scale down
        L0 = Conv3D(filters=number_of_filters,
                    kernel_size=kernel_size,
                    strides=strides,
                    kernel_initializer='glorot_uniform',
                    padding='same')(H0)
        L0 = PReLU(alpha_initializer='zero',
                   shared_axes=[1, 2, 3])(L0)

        # Residual
        E = Subtract()([L0, L])

        # Scale residual up
        H1 = Conv3DTranspose(filters=number_of_filters,
                             kernel_size=kernel_size,
                             strides=strides,
                             kernel_initializer='glorot_uniform',
                             padding='same')(E)
        H1 = PReLU(alpha_initializer='zero',
                   shared_axes=[1, 2, 3])(H1)

        # Output feature map
        up_block = Add()([H0, H1])

        return(up_block)


    def down_block_3d(H, number_of_filters=64, kernel_size=(12, 12, 12), strides=(8, 8, 8),
                    include_dense_convolution_layer=True):
        if include_dense_convolution_layer == True:
            H = Conv3D(filters = number_of_filters,
                       use_bias=True,
                       kernel_size=(1, 1, 1),
                       strides=(1, 1, 1),
                       padding='same')(H)
            H = PReLU(alpha_initializer='zero',
                      shared_axes=[1, 2, 3])(H)

        # Scale down
        L0 = Conv3D(filters=number_of_filters,
                    kernel_size=kernel_size,
                    strides=strides,
                    kernel_initializer='glorot_uniform',
                    padding='same')(H)
        L0 = PReLU(alpha_initializer='zero',
                   shared_axes=[1, 2, 3])(L0)

        # Scale up
        H0 = Conv3DTranspose(filters=number_of_filters,
                             kernel_size=kernel_size,
                             strides=strides,
                             kernel_initializer='glorot_uniform',
                             padding='same')(L0)
        H0 = PReLU(alpha_initializer='zero',
                   shared_axes=[1, 2, 3])(H0)

        # Residual
        E = Subtract()([H0, H])

        # Scale residual down
        L1 = Conv3D(filters=number_of_filters,
                    kernel_size=kernel_size,
                    strides=strides,
                    kernel_initializer='glorot_uniform',
                    padding='same')(E)
        L1 = PReLU(alpha_initializer='zero',
                   shared_axes=[1, 2, 3])(L1)

        # Output feature map
        down_block = Add()([L0, L1])

        return(down_block)

    inputs = Input(shape=input_image_size)

    # Initial feature extraction
    model = Conv3D(filters=number_of_feature_filters,
                   kernel_size=(3, 3, 3),
                   strides=(1, 1, 1),
                   padding='same',
                   kernel_initializer='glorot_uniform')(inputs)
    model = PReLU(alpha_initializer='zero',
                  shared_axes=[1, 2, 3])(model)

    # Feature smashing
    model = Conv3D(filters=number_of_base_filters,
                   kernel_size=(1, 1, 1),
                   strides=(1, 1, 1),
                   padding='same',
                   kernel_initializer='glorot_uniform')(model)
    model = PReLU(alpha_initializer='zero',
                  shared_axes=[1, 2, 3])(model)

    # Back projection
    up_projection_blocks = []
    down_projection_blocks = []

    model = up_block_3d(model, number_of_filters=number_of_base_filters,
      kernel_size=convolution_kernel_size, strides=strides)
    up_projection_blocks.append(model)

    for i in range(number_of_back_projection_stages):
        if i == 0:
            model = down_block_3d(model, number_of_filters=number_of_base_filters,
              kernel_size=convolution_kernel_size, strides=strides)
            down_projection_blocks.append(model)

            model = up_block_3d(model, number_of_filters=number_of_base_filters,
              kernel_size=convolution_kernel_size, strides=strides)
            up_projection_blocks.append(model)

            model = Concatenate()(up_projection_blocks)
        else:
            model = down_block_3d(model, number_of_filters=number_of_base_filters,
              kernel_size=convolution_kernel_size, strides=strides,
              include_dense_convolution_layer=True)
            down_projection_blocks.append(model)
            model = Concatenate()(down_projection_blocks)

            model = up_block_3d(model, number_of_filters=number_of_base_filters,
              kernel_size=convolution_kernel_size, strides=strides,
              include_dense_convolution_layer=True)
            up_projection_blocks.append(model)

            model = Concatenate()(up_projection_blocks)

    # Final convolution layer
    outputs = Conv3D(filters=number_of_outputs,
                     kernel_size=last_convolution,
                     strides=(1, 1, 1),
                     padding = 'same',
                     kernel_initializer = "glorot_uniform")(model)

    if number_of_loss_functions == 1:
        deep_back_projection_network_model = Model(inputs=inputs, outputs=outputs)
    else:
        outputList=[]
        for k in range(number_of_loss_functions):
            outputList.append(outputs)
        deep_back_projection_network_model = Model(inputs=inputs, outputs=outputList)

    return(deep_back_projection_network_model)






























from tensorflow.keras.models import Model
from tensorflow.keras.layers import (Input, Conv2D, Conv3D)

def create_image_super_resolution_model_2d(input_image_size,
                                           convolution_kernel_sizes=[(9, 9), (1, 1), (5, 5)],
                                           number_of_filters=(64, 32)
                                          ):
    """
    2-D implementation of the image super resolution deep learning architecture.

    Creates a keras model of the image super resolution deep learning framework.
    based on the paper available here:

            https://arxiv.org/pdf/1501.00092

    This particular implementation is based on the following python
    implementation:

            https://github.com/titu1994/Image-Super-Resolution

    Arguments
    ---------
    input_image_size : tuple of length 3
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    convolution_kernel_sizes : list of 2-d tuples
        specifies the kernel size at each convolution layer.  Default values are
        the same as given in the original paper.  The length of kernel size list
        must be 1 greater than the tuple length of the number of filters.

    number_of_filters : tuple
       Contains the number of filters for each convolutional layer.  Default values
       are the same as given in the original paper.

    Returns
    -------
    Keras model
        A 2-D Keras model defining the network.

    Example
    -------
    >>> model = create_image_super_resolution_model_2d((128, 128, 1))
    >>> model.summary()
    """

    number_of_convolution_layers = len(convolution_kernel_sizes)

    if len(number_of_filters) != (number_of_convolution_layers - 1):
        raise ValueError("The length of the number of filters must be 1 less than the length of the convolution vector size")

    inputs = Input(shape = input_image_size)

    outputs = inputs

    for i in range(number_of_convolution_layers - 1):
        outputs = Conv2D(filters=number_of_filters[i],
                         kernel_size=convolution_kernel_sizes[i],
                         activation='relu',
                         padding='same')(outputs)

    number_of_channels = input_image_size[-1]

    outputs = Conv2D(filters=number_of_channels,
                     kernel_size=convolution_kernel_sizes[-1],
                     activation='relu',
                     padding='same')(outputs)

    sr_model = Model(inputs=inputs, outputs=outputs)

    return(sr_model)


def create_image_super_resolution_model_3d(input_image_size,
                                           convolution_kernel_sizes=[(9, 9, 9), (1, 1, 1), (5, 5, 5)],
                                           number_of_filters=(64, 32)
                                          ):
    """
    3-D implementation of the image super resolution deep learning architecture.

    Creates a keras model of the image super resolution deep learning framework.
    based on the paper available here:

            https://arxiv.org/pdf/1501.00092

    This particular implementation is based on the following python
    implementation:

            https://github.com/titu1994/Image-Super-Resolution

    Arguments
    ---------
    input_image_size : tuple of length 4
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    convolution_kernel_sizes : list of 3-d tuples
        specifies the kernel size at each convolution layer.  Default values are
        the same as given in the original paper.  The length of kernel size list
        must be 1 greater than the tuple length of the number of filters.

    number_of_filters : tuple
       Contains the number of filters for each convolutional layer.  Default values
       are the same as given in the original paper.

    Returns
    -------
    Keras model
        A 3-D Keras model defining the network.

    Example
    -------
    >>> model = create_image_super_resolution_model_3d((128, 128, 128, 1))
    >>> model.summary()
    """

    number_of_convolution_layers = len(convolution_kernel_sizes)

    if len(number_of_filters) != (number_of_convolution_layers - 1):
        raise ValueError("The length of the number of filters must be 1 less than the length of the convolution vector size")

    inputs = Input(shape = input_image_size)

    outputs = inputs

    for i in range(number_of_convolution_layers - 1):
        outputs = Conv3D(filters=number_of_filters[i],
                         kernel_size=convolution_kernel_sizes[i],
                         activation='relu',
                         padding='same')(outputs)

    number_of_channels = input_image_size[-1]

    outputs = Conv3D(filters=number_of_channels,
                     kernel_size=convolution_kernel_sizes[-1],
                     activation='relu',
                     padding='same')(outputs)

    sr_model = Model(inputs=inputs, outputs=outputs)

    return(sr_model)



import numpy as np
import tensorflow as tf
import ants

def desikan_killiany_tourville_labeling(t1,
                                        do_preprocessing=True,
                                        return_probability_images=False,
                                        do_lobar_parcellation=False,
                                        antsxnet_cache_directory=None,
                                        verbose=False):

    """
    Cortical and deep gray matter labeling using Desikan-Killiany-Tourville

    Perform DKT labeling using deep learning

    The labeling is as follows:

    Inner labels:
    Label 0: background
    Label 4: left lateral ventricle
    Label 5: left inferior lateral ventricle
    Label 6: left cerebellem exterior
    Label 7: left cerebellum white matter
    Label 10: left thalamus proper
    Label 11: left caudate
    Label 12: left putamen
    Label 13: left pallidium
    Label 15: 4th ventricle
    Label 16: brain stem
    Label 17: left hippocampus
    Label 18: left amygdala
    Label 24: CSF
    Label 25: left lesion
    Label 26: left accumbens area
    Label 28: left ventral DC
    Label 30: left vessel
    Label 43: right lateral ventricle
    Label 44: right inferior lateral ventricle
    Label 45: right cerebellum exterior
    Label 46: right cerebellum white matter
    Label 49: right thalamus proper
    Label 50: right caudate
    Label 51: right putamen
    Label 52: right palladium
    Label 53: right hippocampus
    Label 54: right amygdala
    Label 57: right lesion
    Label 58: right accumbens area
    Label 60: right ventral DC
    Label 62: right vessel
    Label 72: 5th ventricle
    Label 85: optic chasm
    Label 91: left basal forebrain
    Label 92: right basal forebrain
    Label 630: cerebellar vermal lobules I-V
    Label 631: cerebellar vermal lobules VI-VII
    Label 632: cerebellar vermal lobules VIII-X

    Outer labels:
    Label 1002: left caudal anterior cingulate
    Label 1003: left caudal middle frontal
    Label 1005: left cuneus
    Label 1006: left entorhinal
    Label 1007: left fusiform
    Label 1008: left inferior parietal
    Label 1009: left inferior temporal
    Label 1010: left isthmus cingulate
    Label 1011: left lateral occipital
    Label 1012: left lateral orbitofrontal
    Label 1013: left lingual
    Label 1014: left medial orbitofrontal
    Label 1015: left middle temporal
    Label 1016: left parahippocampal
    Label 1017: left paracentral
    Label 1018: left pars opercularis
    Label 1019: left pars orbitalis
    Label 1020: left pars triangularis
    Label 1021: left pericalcarine
    Label 1022: left postcentral
    Label 1023: left posterior cingulate
    Label 1024: left precentral
    Label 1025: left precuneus
    Label 1026: left rostral anterior cingulate
    Label 1027: left rostral middle frontal
    Label 1028: left superior frontal
    Label 1029: left superior parietal
    Label 1030: left superior temporal
    Label 1031: left supramarginal
    Label 1034: left transverse temporal
    Label 1035: left insula
    Label 2002: right caudal anterior cingulate
    Label 2003: right caudal middle frontal
    Label 2005: right cuneus
    Label 2006: right entorhinal
    Label 2007: right fusiform
    Label 2008: right inferior parietal
    Label 2009: right inferior temporal
    Label 2010: right isthmus cingulate
    Label 2011: right lateral occipital
    Label 2012: right lateral orbitofrontal
    Label 2013: right lingual
    Label 2014: right medial orbitofrontal
    Label 2015: right middle temporal
    Label 2016: right parahippocampal
    Label 2017: right paracentral
    Label 2018: right pars opercularis
    Label 2019: right pars orbitalis
    Label 2020: right pars triangularis
    Label 2021: right pericalcarine
    Label 2022: right postcentral
    Label 2023: right posterior cingulate
    Label 2024: right precentral
    Label 2025: right precuneus
    Label 2026: right rostral anterior cingulate
    Label 2027: right rostral middle frontal
    Label 2028: right superior frontal
    Label 2029: right superior parietal
    Label 2030: right superior temporal
    Label 2031: right supramarginal
    Label 2034: right transverse temporal
    Label 2035: right insula

    Performing the lobar parcellation is based on the FreeSurfer division
    described here:

    See https://surfer.nmr.mgh.harvard.edu/fswiki/CorticalParcellation

    Frontal lobe:
    Label 1002:  left caudal anterior cingulate
    Label 1003:  left caudal middle frontal
    Label 1012:  left lateral orbitofrontal
    Label 1014:  left medial orbitofrontal
    Label 1017:  left paracentral
    Label 1018:  left pars opercularis
    Label 1019:  left pars orbitalis
    Label 1020:  left pars triangularis
    Label 1024:  left precentral
    Label 1026:  left rostral anterior cingulate
    Label 1027:  left rostral middle frontal
    Label 1028:  left superior frontal
    Label 2002:  right caudal anterior cingulate
    Label 2003:  right caudal middle frontal
    Label 2012:  right lateral orbitofrontal
    Label 2014:  right medial orbitofrontal
    Label 2017:  right paracentral
    Label 2018:  right pars opercularis
    Label 2019:  right pars orbitalis
    Label 2020:  right pars triangularis
    Label 2024:  right precentral
    Label 2026:  right rostral anterior cingulate
    Label 2027:  right rostral middle frontal
    Label 2028:  right superior frontal

    Parietal:
    Label 1008:  left inferior parietal
    Label 1010:  left isthmus cingulate
    Label 1022:  left postcentral
    Label 1023:  left posterior cingulate
    Label 1025:  left precuneus
    Label 1029:  left superior parietal
    Label 1031:  left supramarginal
    Label 2008:  right inferior parietal
    Label 2010:  right isthmus cingulate
    Label 2022:  right postcentral
    Label 2023:  right posterior cingulate
    Label 2025:  right precuneus
    Label 2029:  right superior parietal
    Label 2031:  right supramarginal

    Temporal:
    Label 1006:  left entorhinal
    Label 1007:  left fusiform
    Label 1009:  left inferior temporal
    Label 1015:  left middle temporal
    Label 1016:  left parahippocampal
    Label 1030:  left superior temporal
    Label 1034:  left transverse temporal
    Label 2006:  right entorhinal
    Label 2007:  right fusiform
    Label 2009:  right inferior temporal
    Label 2015:  right middle temporal
    Label 2016:  right parahippocampal
    Label 2030:  right superior temporal
    Label 2034:  right transverse temporal

    Occipital:
    Label 1005:  left cuneus
    Label 1011:  left lateral occipital
    Label 1013:  left lingual
    Label 1021:  left pericalcarine
    Label 2005:  right cuneus
    Label 2011:  right lateral occipital
    Label 2013:  right lingual
    Label 2021:  right pericalcarine

    Other outer labels:
    Label 1035:  left insula
    Label 2035:  right insula

    Preprocessing on the training data consisted of:
       * n4 bias correction,
       * denoising,
       * brain extraction, and
       * affine registration to MNI.
    The input T1 should undergo the same steps.  If the input T1 is the raw
    T1, these steps can be performed by the internal preprocessing, i.e. set
    do_preprocessing = True

    Arguments
    ---------
    t1 : ANTsImage
        raw or preprocessed 3-D T1-weighted brain image.

    do_preprocessing : boolean
        See description above.

    return_probability_images : boolean
        Whether to return the two sets of probability images for the inner and outer
        labels.

    do_lobar_parcellation : boolean
        Perform lobar parcellation (also divided by hemisphere).

    antsxnet_cache_directory : string
        Destination directory for storing the downloaded template and model weights.
        Since these can be resused, if is None, these data will be downloaded to a
        ~/.keras/ANTsXNet/.

    verbose : boolean
        Print progress to the screen.

    Returns
    -------
    List consisting of the segmentation image and probability images for
    each label.

    Example
    -------
    >>> image = ants.image_read("t1.nii.gz")
    >>> dkt = desikan_killiany_tourville_labeling(image)
    """

    from ..architectures import create_unet_model_3d
    from ..utilities import get_pretrained_network
    from ..utilities import get_antsxnet_data
    from ..utilities import preprocess_brain_image
    from ..utilities import deep_atropos

    if t1.dimension != 3:
        raise ValueError( "Image dimension must be 3." )

    if antsxnet_cache_directory == None:
        antsxnet_cache_directory = "ANTsXNet"


    template_transform_type = "antsRegistrationSyNQuickRepro[a]"
    ################################
    #
    # Preprocess images
    #
    ################################

    t1_preprocessed = t1
    if do_preprocessing == True:
        t1_preprocessing = preprocess_brain_image(t1,
            truncate_intensity=(0.01, 0.99),
            brain_extraction_modality="t1",
            template="croppedMni152",
            template_transform_type=template_transform_type,
            do_bias_correction=True,
            do_denoising=True,
            antsxnet_cache_directory=antsxnet_cache_directory,
            verbose=verbose)
        t1_preprocessed = t1_preprocessing["preprocessed_image"] * t1_preprocessing['brain_mask']

    ################################
    #
    # Download spatial priors for outer model
    #
    ################################

    spatial_priors_file_name_path = get_antsxnet_data("priorDktLabels",
      antsxnet_cache_directory=antsxnet_cache_directory)
    spatial_priors = ants.image_read(spatial_priors_file_name_path)
    priors_image_list = ants.ndimage_to_list(spatial_priors)

    ################################
    #
    # Build outer model and load weights
    #
    ################################

    template_size = (96, 112, 96)
    labels = (0, 1002, 1003, *tuple(range(1005, 1032)), 1034, 1035,
                 2002, 2003, *tuple(range(2005, 2032)), 2034, 2035)
    channel_size = 1 + len(priors_image_list)

    unet_model = create_unet_model_3d((*template_size, channel_size),
        number_of_outputs = len(labels),
        number_of_layers = 4, number_of_filters_at_base_layer = 16, dropout_rate = 0.0,
        convolution_kernel_size = (3, 3, 3), deconvolution_kernel_size = (2, 2, 2),
        weight_decay = 1e-5, additional_options=("attentionGating"))

    weights_file_name = None
    weights_file_name = get_pretrained_network("dktOuterWithSpatialPriors",
                                               antsxnet_cache_directory=antsxnet_cache_directory)
    unet_model.load_weights(weights_file_name)

    ################################
    #
    # Do prediction and normalize to native space
    #
    ################################

    if verbose == True:
        print("Outer model Prediction.")

    downsampled_image = ants.resample_image(t1_preprocessed, template_size, use_voxels=True, interp_type=0)
    image_array = downsampled_image.numpy()
    image_array = (image_array - image_array.mean()) / image_array.std()

    batchX = np.zeros((1, *template_size, channel_size))
    batchX[0,:,:,:,0] = image_array

    for i in range(len(priors_image_list)):
        resampled_prior_image = ants.resample_image(priors_image_list[i], template_size, use_voxels=True, interp_type=0)
        batchX[0,:,:,:,i+1] = resampled_prior_image.numpy()

    predicted_data = unet_model.predict(batchX, verbose=verbose)

    origin = downsampled_image.origin
    spacing = downsampled_image.spacing
    direction = downsampled_image.direction

    inner_probability_images = list()
    for i in range(len(labels)):
        probability_image = \
            ants.from_numpy(np.squeeze(predicted_data[0, :, :, :, i]),
            origin=origin, spacing=spacing, direction=direction)
        resampled_image = ants.resample_image( probability_image, t1_preprocessed.shape, use_voxels=True, interp_type=0)
        if do_preprocessing == True:
            inner_probability_images.append(ants.apply_transforms(fixed=t1,
                moving=resampled_image,
                transformlist=t1_preprocessing['template_transforms']['invtransforms'],
                whichtoinvert=[True], interpolator="linear", verbose=verbose))
        else:
            inner_probability_images.append(resampled_image)

    image_matrix = ants.image_list_to_matrix(inner_probability_images, t1 * 0 + 1)
    segmentation_matrix = np.argmax(image_matrix, axis=0)
    segmentation_image = ants.matrix_to_images(
        np.expand_dims(segmentation_matrix, axis=0), t1 * 0 + 1)[0]

    dkt_label_image = ants.image_clone(segmentation_image)
    for i in range(len(labels)):
        dkt_label_image[segmentation_image==i] = labels[i]

    ################################
    #
    # Build inner model and load weights
    #
    ################################

    template_size = (160, 192, 160)
    labels = (0, 4, 6, 7, 10, 11, 12, 13, 14, 15, 16, 17, 18, 24, 26,
        28, 30, 43, 44, 45, 46, 49, 50, 51, 52, 53, 54, 58, 60, 91, 92, 630, 631, 632)

    unet_model = create_unet_model_3d((*template_size, 1),
        number_of_outputs = len(labels),
        number_of_layers = 4, number_of_filters_at_base_layer = 8, dropout_rate = 0.0,
        convolution_kernel_size = (3, 3, 3), deconvolution_kernel_size = (2, 2, 2),
        weight_decay = 1e-5, additional_options=("attentionGating"))

    weights_file_name = get_pretrained_network("dktInner", antsxnet_cache_directory=antsxnet_cache_directory)
    unet_model.load_weights(weights_file_name)

    ################################
    #
    # Do prediction and normalize to native space
    #
    ################################

    if verbose == True:
        print("Prediction.")

    cropped_image = ants.crop_indices(t1_preprocessed, (12, 14, 0), (172, 206, 160))

    batchX = np.expand_dims(cropped_image.numpy(), axis=0)
    batchX = np.expand_dims(batchX, axis=-1)
    batchX = (batchX - batchX.mean()) / batchX.std()

    predicted_data = unet_model.predict(batchX, verbose=verbose)

    origin = cropped_image.origin
    spacing = cropped_image.spacing
    direction = cropped_image.direction

    outer_probability_images = list()
    for i in range(len(labels)):
        probability_image = \
            ants.from_numpy(np.squeeze(predicted_data[0, :, :, :, i]),
            origin=origin, spacing=spacing, direction=direction)
        if i > 0:
            decropped_image = ants.decrop_image(probability_image, t1_preprocessed * 0)
        else:
            decropped_image = ants.decrop_image(probability_image, t1_preprocessed * 0 + 1)

        if do_preprocessing == True:
            outer_probability_images.append(ants.apply_transforms(fixed=t1,
                moving=decropped_image,
                transformlist=t1_preprocessing['template_transforms']['invtransforms'],
                whichtoinvert=[True], interpolator="linear", verbose=verbose))
        else:
            outer_probability_images.append(decropped_image)

    image_matrix = ants.image_list_to_matrix(outer_probability_images, t1 * 0 + 1)
    segmentation_matrix = np.argmax(image_matrix, axis=0)
    segmentation_image = ants.matrix_to_images(
        np.expand_dims(segmentation_matrix, axis=0), t1 * 0 + 1)[0]

    ################################
    #
    # Incorporate the inner model results into the final label image.
    # Note that we purposely prioritize the inner label results.
    #
    ################################

    for i in range(len(labels)):
        if labels[i] > 0:
            dkt_label_image[segmentation_image==i] = labels[i]

    if do_lobar_parcellation:

        if verbose == True:
            print("Doing lobar parcellation.")

        ################################
        #
        # Lobar/hemisphere parcellation
        #
        ################################

        # Consolidate lobar cortical labels

        if verbose == True:
            print("   Consolidating cortical labels.")

        frontal_labels = (1002, 1003, 1012, 1014, 1017, 1018, 1019, 1020, 1024, 1026, 1027, 1028,
                            2002, 2003, 2012, 2014, 2017, 2018, 2019, 2020, 2024, 2026, 2027, 2028)
        parietal_labels = (1008, 1010, 1022, 1023, 1025, 1029, 1031,
                            2008, 2010, 2022, 2023, 2025, 2029, 2031)
        temporal_labels = (1006, 1007, 1009, 1015, 1016, 1030, 1034,
                            2006, 2007, 2009, 2015, 2016, 2030, 2034)
        occipital_labels = (1005, 1011, 1013, 1021,
                            2005, 2011, 2013, 2021)

        lobar_labels = list()
        lobar_labels.append(frontal_labels)
        lobar_labels.append(parietal_labels)
        lobar_labels.append(temporal_labels)
        lobar_labels.append(occipital_labels)

        dkt_lobes = ants.image_clone(dkt_label_image)
        dkt_lobes[dkt_lobes < 1000] = 0

        for i in range(len(lobar_labels)):
            for j in range(len(lobar_labels[i])):
                dkt_lobes[dkt_lobes == lobar_labels[i][j]] = i + 1

        dkt_lobes[dkt_lobes > len(lobar_labels)] = 0

        six_tissue = deep_atropos(t1_preprocessed, do_preprocessing=False,
            antsxnet_cache_directory=antsxnet_cache_directory, verbose=verbose)
        atropos_seg = six_tissue['segmentation_image']
        if do_preprocessing == True:
            atropos_seg = ants.apply_transforms(fixed=t1, moving=atropos_seg,
                transformlist=t1_preprocessing['template_transforms']['invtransforms'],
                whichtoinvert=[True], interpolator="genericLabel", verbose=verbose)

        brain_mask = ants.image_clone(atropos_seg)
        brain_mask[brain_mask == 1 or brain_mask == 5 or brain_mask == 6] = 0
        brain_mask = ants.threshold_image(brain_mask, 0, 0, 0, 1)

        lobar_parcellation = ants.iMath(brain_mask, "PropagateLabelsThroughMask", brain_mask * dkt_lobes)

        lobar_parcellation[atropos_seg == 5] = 5
        lobar_parcellation[atropos_seg == 6] = 6

        # Do left/right

        if verbose == True:
            print("   Doing left/right hemispheres.")


        left_labels = (*tuple(range(4, 8)), *tuple(range(10, 14)), 17, 18, 25, 26, 28, 30, 91,
                       1002, 1003, *tuple(range(1005, 1032)), 1034, 1035)
        right_labels = (*tuple(range(43, 47)), *tuple(range(49, 55)), 57, 58, 60, 62, 92, 2002, 2003,
                        *tuple(range(2005, 2032)), 2034, 2035)

        hemisphere_labels = list()
        hemisphere_labels.append(left_labels)
        hemisphere_labels.append(right_labels)

        dkt_hemispheres = ants.image_clone(dkt_label_image)

        for i in range(len(hemisphere_labels)):
            for j in range(len(hemisphere_labels[i])):
                dkt_hemispheres[dkt_hemispheres == hemisphere_labels[i][j]] = i + 1

        dkt_hemispheres[dkt_hemispheres > 2] = 0

        atropos_brain_mask = ants.threshold_image(atropos_seg, 0, 0, 0, 1)
        hemisphere_parcellation = ants.iMath(atropos_brain_mask, "PropagateLabelsThroughMask",
          atropos_brain_mask * dkt_hemispheres)

        # The following contains a bug somewhere as only the latter condition is seen.
        # Need to fix it.
        #
        # for i in range(6):
        #     lobar_parcellation[lobar_parcellation == (i + 1) and hemisphere_parcellation == 2] = 6 + i + 1

        hemisphere_parcellation *= ants.threshold_image(lobar_parcellation, 0, 0, 0, 1)
        hemisphere_parcellation[hemisphere_parcellation == 1] = 0
        hemisphere_parcellation[hemisphere_parcellation == 2] = 1
        hemisphere_parcellation *= 6
        lobar_parcellation += hemisphere_parcellation

    if return_probability_images == True and do_lobar_parcellation == True:
        return_dict = {'segmentation_image' : dkt_label_image,
                       'lobar_parcellation' : lobar_parcellation,
                       'inner_probability_images' : inner_probability_images,
                       'outer_probability_images' : outer_probability_images }
        return(return_dict)
    elif return_probability_images == True and do_lobar_parcellation == False:
        return_dict = {'segmentation_image' : dkt_label_image,
                       'inner_probability_images' : inner_probability_images,
                       'outer_probability_images' : outer_probability_images }
        return(return_dict)
    elif return_probability_images == False and do_lobar_parcellation == True:
        return_dict = {'segmentation_image' : dkt_label_image,
                       'lobar_parcellation' : lobar_parcellation }
        return(return_dict)
    else:
        return(dkt_label_image)



import ants

import numpy as np

def histogram_warp_image_intensities(image,
                                     break_points=(0.25, 0.5, 0.75),
                                     displacements=None,
                                     clamp_end_points=(False, False),
                                     sd_displacements=0.05,
                                     transform_domain_size=20):
    """
    Transform image intensities based on histogram mapping.

    Apply B-spline 1-D maps to an input image for intensity warping.

    Arguments
    ---------
    image : ANTsImage
        Input image.

    break_points : integer or tuple
        Parametric points at which the intensity transform displacements
        are specified between [0, 1].  Alternatively, a single number can
        be given and the sequence is linearly spaced in [0, 1].

    displacements : tuple
        displacements to define intensity warping.  Length must be equal to the
        breakPoints.  Alternatively, if None random displacements are chosen
        (random normal:  mean = 0, sd = sd_displacements).

    sd_displacements : float
        Characterize the randomness of the intensity displacement.

    clamp_end_points : 2-element tuple of booleans
        Specify non-zero intensity change at the ends of the histogram.

    transform_domain_size : integer
        Defines the sampling resolution of the B-spline warping.

    Returns
    -------
    ANTs image

    Example
    -------
    >>> import ants
    >>> image = ants.image_read(ants.get_ants_data("r64"))
    >>> transformed_image = histogram_warp_image_intensities( image )
    """

    if not len(clamp_end_points) == 2:
        raise ValueError("clamp_end_points must be a boolean tuple of length 2.")

    if not isinstance(break_points, int):
        if any(b < 0 for b in break_points) and any(b > 1 for b in break_points):
            raise ValueError("If specifying break_points as a vector, values must be in the range [0, 1]")

    parametric_points = None
    number_of_nonzero_displacements = 1
    if not isinstance(break_points, int):
        parametric_points = break_points
        number_of_nonzero_displacements = len(break_points)
        if clamp_end_points[0] is True:
            parametric_points = (0, *parametric_points)
        if clamp_end_points[1] is True:
            parametric_points = (*parametric_points, 1)
    else:
        total_number_of_break_points = break_points
        if clamp_end_points[0] is True:
            total_number_of_break_points += 1
        if clamp_end_points[1] is True:
            total_number_of_break_points += 1
        parametric_points = np.linspace(0, 1, total_number_of_break_points)
        number_of_nonzero_displacements = break_points

    if displacements is None:
        displacements = np.random.normal(loc=0.0, scale=sd_displacements, size=number_of_nonzero_displacements)

    weights = np.ones(len(displacements))
    if clamp_end_points[0] is True:
        displacements = (0, *displacements)
        weights = np.concatenate((1000 * np.ones(1), weights))
    if clamp_end_points[1] is True:
        displacements = (*displacements, 0)
        weights = np.concatenate((weights, 1000 * np.ones(1)))

    if not len(displacements) == len(parametric_points):
        raise ValueError("Length of displacements does not match the length of the break points.")

    scattered_data = np.reshape(displacements, (len(displacements), 1))
    parametric_data = np.reshape(parametric_points, (len(parametric_points), 1))

    transform_domain_origin = 0
    transform_domain_spacing = (1.0 - transform_domain_origin) / (transform_domain_size - 1)

    bspline_histogram_transform = ants.fit_bspline_object_to_scattered_data( scattered_data,
        parametric_data, [transform_domain_origin], [transform_domain_spacing], [transform_domain_size],
        data_weights=weights)

    transform_domain = np.linspace(0, 1, transform_domain_size)

    normalized_image = ants.iMath(ants.image_clone(image), "Normalize")
    transformed_array = normalized_image.numpy()
    normalized_array = normalized_image.numpy()

    for i in range(len(transform_domain) - 1):
        indices = np.where((normalized_array >= transform_domain[i]) & (normalized_array < transform_domain[i+1]))
        intensities = normalized_array[indices]

        alpha = (intensities - transform_domain[i])/(transform_domain[i+1] - transform_domain[i])
        xfrm = alpha * (bspline_histogram_transform[i+1] - bspline_histogram_transform[i]) + bspline_histogram_transform[i]
        transformed_array[indices] = intensities + xfrm

    transformed_image = (ants.from_numpy(transformed_array, origin=image.origin,
        spacing=image.spacing, direction=image.direction) * (image.max() - image.min())) + image.min()

    return(transformed_image)





from tensorflow.keras.models import Model
from tensorflow.keras.layers import (Input, Flatten, Dense,
                          Conv2D, Conv2DTranspose, MaxPooling2D,
                          ZeroPadding2D,
                          Conv3D, Conv3DTranspose, MaxPooling3D,
                          ZeroPadding3D)

def create_vgg_model_2d(input_image_size,
                        number_of_classification_labels=1000,
                        layers=(1, 2, 3, 4, 4),
                        lowest_resolution=64,
                        convolution_kernel_size=(3, 3),
                        pool_size=(2, 2),
                        strides=(2, 2),
                        number_of_dense_units=4096,
                        dropout_rate=0.0,
                        style=19,
                        mode='classification'):
    """
    2-D implementation of the Vgg deep learning architecture.

    Creates a keras model of the Vgg deep learning architecture for image
    recognition based on the paper

    K. Simonyan and A. Zisserman, Very Deep Convolutional Networks for
    Large-Scale Image Recognition

    available here:

            https://arxiv.org/abs/1409.1556

    This particular implementation was influenced by the following python
    implementation:

            https://gist.github.com/baraldilorenzo/8d096f48a1be4a2d660d

    Arguments
    ---------

    input_image_size : tuple of length 3
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    number_of_classification_labels : integer
        Number of classification labels.

    layers : tuple
        A tuple determining the number of 'filters' defined at for each layer.

    lowest_resolution : integer
        Number of filters at the beginning.

    convolution_kernel_size : tuple
        2-d vector definining the kernel size during the encoding path

    pool_size : tuple
        2-d vector defining the region for each pooling layer.

    strides : tuple
        2-d vector describing the stride length in each direction.

    number_of_dense_units : integer
        Number of units in the last layers.

    dropout_rate : scalar
       Between 0 and 1 to use between dense layers.

    style : integer
       '16' or '19' for VGG16 or VGG19, respectively.

    mode : string
        'classification' or 'regression'.  Default = 'classification'.

    Returns
    -------
    Keras model
        A 2-D Keras model defining the network.

    Example
    -------
    >>> model = create_vgg_model_2d((128, 128, 1))
    >>> model.summary()

    """

    if style != 16 and style != 19:
        raise ValueError("Incorrect style.  Must be either '16' or '19'.")

    inputs = Input(shape = input_image_size)
    outputs = None

    for i in range(len(layers)):
        number_of_filters = lowest_resolution * 2**(layers[i] - 1)

        if i == 0:
            outputs = Conv2D(filters=number_of_filters,
                             kernel_size=convolution_kernel_size,
                             activation='relu',
                             padding='same')(inputs)
            outputs = Conv2D(filters=number_of_filters,
                             kernel_size=convolution_kernel_size,
                             activation='relu',
                             padding='same')(outputs)
        elif i == 1:
            outputs = Conv2D(filters=number_of_filters,
                             kernel_size=convolution_kernel_size,
                             activation='relu',
                             padding='same')(outputs)
            outputs = Conv2D(filters=number_of_filters,
                             kernel_size=convolution_kernel_size,
                             activation='relu',
                             padding='same')(outputs)
        else:
            outputs = Conv2D(filters=number_of_filters,
                             kernel_size=convolution_kernel_size,
                             activation='relu',
                             padding='same')(outputs)
            outputs = Conv2D(filters=number_of_filters,
                             kernel_size=convolution_kernel_size,
                             activation='relu',
                             padding='same')(outputs)
            outputs = Conv2D(filters=number_of_filters,
                             kernel_size=convolution_kernel_size,
                             activation='relu',
                             padding='same')(outputs)
            if style == 19:
                outputs = Conv2D(filters=number_of_filters,
                                 kernel_size=convolution_kernel_size,
                                 activation='relu',
                                 padding='same')(outputs)
        outputs = MaxPooling2D(pool_size=pool_size,
                               strides=strides)(outputs)

    outputs = Flatten()(outputs)
    outputs = Dense(units=number_of_dense_units,
                    activation ='relu')(outputs)
    if dropout_rate > 0.0:
        outputs = Dropout(rate=dropout_rate)(outputs)
    outputs = Dense(units=number_of_dense_units,
                    activation ='relu')(outputs)
    if dropout_rate > 0.0:
        outputs = Dropout(rate=dropout_rate)(outputs)

    layer_activation = ''
    if mode == 'classification':
        layer_activation = 'softmax'
    elif mode == 'regression':
        layerActivation = 'linear'
    else:
        raise ValueError('unrecognized mode.')

    outputs = Dense(units=number_of_classification_labels,
                    activation =layer_activation)(outputs)

    vgg_model = Model(inputs=inputs, outputs=outputs)

    return(vgg_model)

def create_vgg_model_3d(input_image_size,
                        number_of_classification_labels=1000,
                        layers=(1, 2, 3, 4, 4),
                        lowest_resolution=64,
                        convolution_kernel_size=(3, 3, 3),
                        pool_size=(2, 2, 2),
                        strides=(2, 2, 2),
                        number_of_dense_units=4096,
                        dropout_rate=0.0,
                        style=19,
                        mode='classification'):
    """
    3-D implementation of the Vgg deep learning architecture.

    Creates a keras model of the Vgg deep learning architecture for image
    recognition based on the paper

    K. Simonyan and A. Zisserman, Very Deep Convolutional Networks for
    Large-Scale Image Recognition

    available here:

            https://arxiv.org/abs/1409.1556

    This particular implementation was influenced by the following python
    implementation:

            https://gist.github.com/baraldilorenzo/8d096f48a1be4a2d660d

    Arguments
    ---------

    input_image_size : tuple of length 4
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    number_of_classification_labels : integer
        Number of classification labels.

    layers : tuple
        A tuple determining the number of 'filters' defined at for each layer.

    lowest_resolution : integer
        Number of filters at the beginning.

    convolution_kernel_size : tuple
        3-d vector definining the kernel size during the encoding path

    pool_size : tuple
        3-d vector defining the region for each pooling layer.

    strides : tuple
        3-d vector describing the stride length in each direction.

    number_of_dense_units : integer
        Number of units in the last layers.

    dropout_rate : scalar
       Between 0 and 1 to use between dense layers.

    style : integer
       '16' or '19' for VGG16 or VGG19, respectively.

    mode : string
        'classification' or 'regression'.  Default = 'classification'.

    Returns
    -------
    Keras model
        A 3-D Keras model defining the network.

    Example
    -------
    >>> model = create_vgg_model_3d((128, 128, 128, 1))
    >>> model.summary()

    """

    if style != 16 and style != 19:
        raise ValueError("Incorrect style.  Must be either '16' or '19'.")

    inputs = Input(shape = input_image_size)
    outputs = None

    for i in range(len(layers)):
        number_of_filters = lowest_resolution * 2**(layers[i] - 1)

        if i == 0:
            outputs = Conv3D(filters=number_of_filters,
                             kernel_size=convolution_kernel_size,
                             activation='relu',
                             padding='same')(inputs)
            outputs = Conv3D(filters=number_of_filters,
                             kernel_size=convolution_kernel_size,
                             activation='relu',
                             padding='same')(outputs)
        elif i == 1:
            outputs = Conv3D(filters=number_of_filters,
                             kernel_size=convolution_kernel_size,
                             activation='relu',
                             padding='same')(outputs)
            outputs = Conv3D(filters=number_of_filters,
                             kernel_size=convolution_kernel_size,
                             activation='relu',
                             padding='same')(outputs)
        else:
            outputs = Conv3D(filters=number_of_filters,
                             kernel_size=convolution_kernel_size,
                             activation='relu',
                             padding='same')(outputs)
            outputs = Conv3D(filters=number_of_filters,
                             kernel_size=convolution_kernel_size,
                             activation='relu',
                             padding='same')(outputs)
            outputs = Conv3D(filters=number_of_filters,
                             kernel_size=convolution_kernel_size,
                             activation='relu',
                             padding='same')(outputs)
            if style == 19:
                outputs = Conv3D(filters=number_of_filters,
                                 kernel_size=convolution_kernel_size,
                                 activation='relu',
                                 padding='same')(outputs)
        outputs = MaxPooling3D(pool_size=pool_size,
                               strides=strides)(outputs)

    outputs = Flatten()(outputs)
    outputs = Dense(units=number_of_dense_units,
                    activation ='relu')(outputs)
    if dropout_rate > 0.0:
        outputs = Dropout(rate=dropout_rate)(outputs)
    outputs = Dense(units=number_of_dense_units,
                    activation ='relu')(outputs)
    if dropout_rate > 0.0:
        outputs = Dropout(rate=dropout_rate)(outputs)

    layer_activation = ''
    if mode == 'classification':
        layer_activation = 'softmax'
    elif mode == 'regression':
        layerActivation = 'linear'
    else:
        raise ValueError('unrecognized mode.')

    outputs = Dense(units=number_of_classification_labels,
                    activation =layer_activation)(outputs)

    vgg_model = Model(inputs=inputs, outputs=outputs)

    return(vgg_model)





import tensorflow.keras.backend as K
from tensorflow.keras.layers import Layer, InputSpec, Concatenate
from tensorflow.keras import initializers

import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp

class MixtureDensityLayer(Layer):

    """
    Layer for modeling arbitrary functions using neural networks.

    Arguments
    ---------
    output_dimension : integer
        Dimensionality of the output.

    number_of_mixtures : integer
        Number of gaussians used.

    Returns
    -------
    Layer
        A keras layer

    """

    def __init__(self, output_dimension, number_of_mixtures, **kwargs):
        if K.backend() != 'tensorflow':
            raise ValueError("Tensorflow required as the backend.")

        self.output_dimension = output_dimension
        self.number_of_mixtures = number_of_mixtures

        super(MixtureDensityLayer, self).__init__(**kwargs)

    def build(self, input_shape):

        input_dimension = input_shape[-1]

        units1 = self.output_dimension * self.number_of_mixtures

        self.mu_kernel = self.add_weight(name="mu_kernel",
                                         shape = shape(input_dimension, units1),
                                         initializer=initializers.random_normal(),
                                         trainable=True)
        self.mu_bias = self.add_weight(name="mu_bias",
                                       shape = shape(units1),
                                       initializer=initializers.zeros(),
                                       trainable=True)

        self.sigma_kernel = self.add_weight(name="sigma_kernel",
                                            shape = shape(input_dimension, units1),
                                            initializer=initializers.random_normal(),
                                            trainable=True)
        self.sigma_bias = self.add_weight(name="sigma_bias",
                                          shape = shape(units1),
                                          initializer=initializers.zeros(),
                                          trainable=True)

        units2 = self.number_of_mixtures

        self.pi_kernel = self.add_weight(name="pi_kernel",
                                         shape = shape(input_dimension, units2),
                                         initializer=initializers.random_normal(),
                                         trainable=True)
        self.pi_bias = self.add_weight(name="pi_bias",
                                       shape = shape(units2),
                                       initializer=initializers.zeros(),
                                       trainable=True)

    def call(self, inputs, mask=None):

        # dense layer for mu (mean) of the gaussians
        mu_output = K.dot(inputs, self.mu_kernel)
        mu_output = K.bias_add(mu_output, self.mu_bias, data_format='channels_last')

        # dense layer for sigma (variance) of the gaussians
        sigma_output = K.dot(inputs, self.sigma_kernel)
        sigma_output = K.bias_add(sigma_output, self.sigma_bias, data_format='channels_last')

        # Avoid NaN's by pushing sigma through the following custom activation
        sigma_output = K.elu(sigma_output) + 1 + K.epsilon()

        # dense layer for pi (amplitude) of the gaussians
        pi_output = K.dot( inputs, self.pi_kernel)
        pi_output = K.bias_add(pi_output, self.pi_bias, data_format='channels_last')

        output = Concatenate()([mu_output, sigma_output, pi_output], name="mdn_outputs")
        return(output)

    def compute_output_shape(input_shape):
        units = self.number_of_mixtures * (2 * self.output_dimension + 1)
        return((input_shape[0], units))

    def get_config(self):
        config = {"output_dimension": self.output_dimension,
                  "axis": self.number_of_mixtures}
        base_config = super(MixtureDensityLayer, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

def get_mixture_density_loss_function(output_dimension, number_of_mixtures):

    """
    Returns a loss function for the mixture density.

    Arguments
    ---------
    output_dimension : integer
        Dimensionality of the output.

    number_of_mixtures : integer
        Number of gaussians used.

    Returns
    -------
    Function
        A function providing the mean square error accuracy

    """

    def loss_function(y_true, y_pred):

        dimension = number_of_mixtures * output_dimension

        y_pred = tf.reshape(y_pred, [-1, 2 * dimension + number_of_mixtures],
                            name='reshape_ypred_loss')
        y_true = tf.reshape(y_true, [-1, 2 * dimension + number_of_mixtures],
                            name='reshape_ytrue_loss')

        output_mu, output_sigma, output_pi = tf.split(y_pred, axis=-1, name='mdn_coef_split',
                   num_or_size_splits=[dimension, dimension, number_of_mixtures])

        # Construct the mixture models

        tfd = tfp.distributions

        categorical_distribution = tfd.Categorical(logits=output_pi)
        component_splits = [output_dimension] * number_of_mixtures
        mu = tf.split(output_mu, num_or_size_splits=component_splits, axis=1)
        sigma = tf.split(output_sigma, num_or_size_splits=component_splits, axis=1)

        components = []
        for i in range(len(mu)):
            components.append(tfd.MultivariateNormalDiag(loc = mu[i], scale_diag=sigma[i]))

        mixture = tfd.Mixture(cat=categorical_distribution, components=components)

        loss = mixture.log_prob(y_true)
        loss = tf.negative(loss)
        loss = tf.reduce_mean(loss)

        return(loss)

    with tf.name_scope("MixtureDensityNetwork"):
        return(loss_function)

def get_mixture_density_sampling_function(output_dimension, number_of_mixtures):

    """
    Returns a sampling function for the mixture density.

    Arguments
    ---------
    output_dimension : integer
        Dimensionality of the output.

    number_of_mixtures : integer
        Number of gaussians used.

    Returns
    -------
    Function
        A function providing the mean square error accuracy

    """

    def sampling_function(y_pred):

        dimension = number_of_mixtures * output_dimension

        y_pred = tf.reshape(y_pred, [-1, 2 * dimension + number_of_mixtures],
                            name='reshape_ypred')

        output_mu, output_sigma, output_pi = tf.split(y_pred, axis=-1, name='mdn_coef_split',
                   num_or_size_splits=[dimension, dimension, number_of_mixtures])

        # Construct the mixture models

        tfd = tfp.distributions

        categorical_distribution = tfd.Categorical(logits=output_pi)
        component_splits = [output_dimension] * number_of_mixtures
        mu = tf.split(output_mu, num_or_size_splits=component_splits, axis=1)
        sigma = tf.split(output_sigma, num_or_size_splits=component_splits, axis=1)

        components = []
        for i in range(len(mu)):
            components.append(tfd.MultivariateNormalDiag(loc = mu[i], scale_diag=sigma[i]))

        mixture = tfd.Mixture(cat=categorical_distribution, components=components)

        sample = mixture.sample()

        return(sample)

    with tf.name_scope("MixtureDensityNetwork"):
        return(sampling_function)


def get_mixture_density_mse_function(output_dimension, number_of_mixtures):

    """
    Returns a mse function for the mixture density.

    Arguments
    ---------
    output_dimension : integer
        Dimensionality of the output.

    number_of_mixtures : integer
        Number of gaussians used.

    Returns
    -------
    Function
        A function providing the mean square error accuracy

    """

    def mse_accuracy_function(y_true, y_pred):

        dimension = number_of_mixtures * output_dimension

        y_pred = tf.reshape(y_pred, [-1, 2 * dimension + number_of_mixtures],
                            name='reshape_ypred_mse')
        y_true = tf.reshape(y_true, [-1, output_dimension],
                            name='reshape_ytrue_mse')

        output_mu, output_sigma, output_pi = tf.split(y_pred, axis=-1, name='mdn_coef_split',
                   num_or_size_splits=[dimension, dimension, number_of_mixtures])

        # Construct the mixture models

        tfd = tfp.distributions

        categorical_distribution = tfd.Categorical(logits=output_pi)
        component_splits = [output_dimension] * number_of_mixtures
        mu = tf.split(output_mu, num_or_size_splits=component_splits, axis=1)
        sigma = tf.split(output_sigma, num_or_size_splits=component_splits, axis=1)

        components = []
        for i in range(len(mu)):
            components.append(tfd.MultivariateNormalDiag(loc = mu[i], scale_diag=sigma[i]))

        mixture = tfd.Mixture(cat=categorical_distribution, components=components)

        sample = mixture.sample()
        mse = tf.reduce_mean(tf.square(sample-y_true), axis=-1)

        return(mse)

    with tf.name_scope("MixtureDensityNetwork"):
        return(mse_accuracy_function)

def split_mixture_parameters(parameters, output_dimension, number_of_mixtures):

    """
    Splits the mixture parameters.

    Arguments
    ---------

    parameters : tuple
        Parameter to split

    output_dimension : integer
        Dimensionality of the output.

    number_of_mixtures : integer
        Number of gaussians used.

    Returns
    -------
    List of arrays
        Separate mixture parameters

    """

    dimension = number_of_mixtures * output_dimension
    mu = parameters[:dimension]
    sigma = parameters[dimension:(2 * dimension)]
    pi_logits = parameters[-number_of_mixtures:]
    return([mu, sigma, pi_logits])

def mixture_density_software_max(logits, temperature=1.0):

    """
    Softmax function for mixture density with temperature adjustment.

    Arguments
    ---------

    logits : list or numpy array
        input

    temperature :
        The temperature for to adjust the distribution (default 1.0)

    Returns
    -------
    Scalar
        Softmax loss value.

    """

    e = np.array(logits) / temperature
    e -= np.max(e)
    e = np.exp(e)

    distribution = e / np.sum(e)

    return(distribution)

def sample_from_categorical_distribution(distribution):

    """
    Softmax function for mixture density with temperature adjustment.

    Arguments
    ---------

    distribution :
        input categorical distribution from which to sample.

    Returns
    -------
    Scalar
        A single sample.

    """

    r = np.random.rand(1)

    accumulate = 0
    for i in range(len(distribution)):
        accumulate += distribution[i]
        if accumulate >= r:
            return(i)

    tf.logging.info('Error: sampling categorical model.')
    return(-1)

def sample_from_output(parameters, output_dimension, number_of_mixtures,
                       temperature=1.0, sigma_temperature=1.0):

    """
    Softmax function for mixture density with temperature adjustment.

    Arguments
    ---------
    output_dimension : integer
        Dimensionality of the output.

    number_of_mixtures : integer
        Number of gaussians used.

    temperature :
        The temperature for to adjust the distribution (default 1.0)

    sigma_temperature :
        The temperature for to adjust the distribution (default 1.0)

    Returns
    -------
    Scalar
        A single sample.

    """

    mu, sigma, pi = split_mixture_parameters(parameters, output_dimension, number_of_mixtures)
    pi_softmax = mixture_density_software_max(pi, temperature=temperature)
    m = sample_from_categorical_distribution(pi_softmax)

    mu_vector = mu[m * output_dimension:(m + 1) * output_dimension]
    sigma_vector = sigma[m * output_dimension:(m + 1) * output_dimension] * sigma_temperature
    covariance_matrix = np.identity(output_dimension) * sigma_vector
    sample = np.random.multivariate_normal(mu_vector, covariance_matrix, 1)
    return(sample)




from tensorflow.keras.models import Model
from tensorflow.keras.layers import (Input, Dense, Activation, Flatten,
                          Conv2D, MaxPooling2D,
                          Conv3D, MaxPooling3D)

from ..utilities import (SpatialTransformer2D, SpatialTransformer3D)

import numpy as np
import tensorflow as tf

def create_simple_classification_with_spatial_transformer_network_model_2d(input_image_size,
                                                                           resampled_size=(30, 30),
                                                                           number_of_classification_labels=10):
    """
    2-D implementation of the spatial transformer network.

    Creates a keras model of the spatial transformer network:

            https://arxiv.org/abs/1506.02025

    based on the following python Keras model:

            https://github.com/oarriaga/STN.keras/blob/master/src/models/STN.py

    @param inputImageSize Used for specifying the input tensor shape.  The
    shape (or dimension) of that tensor is the image dimensions followed by
    the number of channels (e.g., red, green, and blue).  The batch size
    (i.e., number of training images) is not specified a priori.
    @param resampledSize resampled size of the transformed input images.
    @param numberOfClassificationLabels Number of classes.

    Arguments
    ---------
    input_image_size : tuple of length 3
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    resampled_size : tuple of length 2
        Resampled size of the transformed input images.

    number_of_classification_labels : integer
        Number of units in the final dense layer.

    Returns
    -------
    Keras model
        A 2-D Keras model defining the network.

    Example
    -------
    >>> model = create_simple_classification_with_spatial_transformer_network_model_2d((128, 128, 1))
    >>> model.summary()
    """

    def get_initial_weights_2d(output_size):
        b = np.zeros((2, 3), dtype='float32')
        b[0, 0] = 1
        b[1, 1] = 1

        W = np.zeros((output_size, 6), dtype='float32')

        weights = [W, b.flatten()]
        return(weights)

    inputs = Input(shape = input_image_size)

    localization = inputs
    localization = MaxPooling2D(pool_size=(2, 2))(localization)
    localization = Conv2D(filters=20,
                          kernel_size=(5, 5))(localization)
    localization = MaxPooling2D(pool_size=(2, 2))(localization)
    localization = Conv2D(filters=20,
                          kernel_size=(5, 5))(localization)

    localization = Flatten()(localization)
    localization = Dense(units=50)(localization)
    localization = Activation('relu')(localization)

    weights = get_initial_weights_2d(output_size=50)
    localization = Dense(6, kernel_initializer = tf.constant_initializer(weights[0]),
                            bias_initializer = tf.constant_initializer(weights[1]))(localization)

    outputs = SpatialTransformer2D(resampled_size=resampled_size,
                                   transform_type="affine",
                                   interpolator_type="linear")([inputs, localization])
    outputs = Conv2D(filters=32,
                     kernel_size=(3, 3),
                     padding='same')(outputs)
    outputs = Activation('relu')(outputs)
    outputs = MaxPooling2D(pool_size=(2, 2))(outputs)
    outputs = Conv2D(filters=32,
                     kernel_size=(3, 3))(outputs)
    outputs = Activation('relu')(outputs)
    outputs = MaxPooling2D(pool_size=(2, 2))(outputs)
    outputs = Flatten()(outputs)
    outputs = Dense(units=256)(outputs)
    outputs = Activation('relu')(outputs)
    outputs = Dense(units=number_of_classification_labels)(outputs)

    outputs = Activation('softmax')(outputs)

    stnModel = Model(inputs=inputs, outputs=outputs)

    return(stnModel)

def create_simple_classification_with_spatial_transformer_network_model_3d(input_image_size,
                                                                           resampled_size=(30, 30, 30),
                                                                           number_of_classification_labels=10):
    """
    3-D implementation of the spatial transformer network.

    Creates a keras model of the spatial transformer network:

            https://arxiv.org/abs/1506.02025

    based on the following python Keras model:

            https://github.com/oarriaga/STN.keras/blob/master/src/models/STN.py

    @param inputImageSize Used for specifying the input tensor shape.  The
    shape (or dimension) of that tensor is the image dimensions followed by
    the number of channels (e.g., red, green, and blue).  The batch size
    (i.e., number of training images) is not specified a priori.
    @param resampledSize resampled size of the transformed input images.
    @param numberOfClassificationLabels Number of classes.

    Arguments
    ---------
    input_image_size : tuple of length 4
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    resampled_size : tuple of length 3
        Resampled size of the transformed input images.

    number_of_classification_labels : integer
        Number of units in the final dense layer.

    Returns
    -------
    Keras model
        A 3-D Keras model defining the network.

    Example
    -------
    >>> model = create_simple_classification_with_spatial_transformer_network_model_3d((128, 128, 128, 1))
    >>> model.summary()
    """

    def get_initial_weights_3d(output_size):
        b = np.zeros((3, 4), dtype='float32')
        b[0, 0] = 1
        b[1, 1] = 1
        b[2, 2] = 1

        W = np.zeros((output_size, 12), dtype='float32')

        weights = [W, b.flatten()]
        return(weights)

    inputs = Input(shape = input_image_size)

    localization = inputs
    localization = MaxPooling3D(pool_size=(2, 2, 2))(localization)
    localization = Conv3D(filters=20,
                          kernel_size=(5, 5, 5))(localization)
    localization = MaxPooling3D(pool_size=(2, 2, 2))(localization)
    localization = Conv3D(filters=20,
                          kernel_size=(5, 5, 5))(localization)

    localization = Flatten()(localization)
    localization = Dense(units=50)(localization)
    localization = Activation('relu')(localization)

    weights = get_initial_weights_3d(output_size=50)
    localization = Dense(6, kernel_initializer = tf.constant_initializer(weights[0]),
                            bias_initializer = tf.constant_initializer(weights[1]))(localization)

    outputs = SpatialTransformer3D(resampled_size=resampled_size,
                                   transform_type="affine",
                                   interpolator_type="linear")([inputs, localization])
    outputs = Conv3D(filters=32,
                     kernel_size=(3, 3, 3),
                     padding='same')(outputs)
    outputs = Activation('relu')(outputs)
    outputs = MaxPooling3D(pool_size=(2, 2, 2))(outputs)
    outputs = Conv3D(filters=32,
                     kernel_size=(3, 3, 3))(outputs)
    outputs = Activation('relu')(outputs)
    outputs = MaxPooling3D(pool_size=(2, 2, 2))(outputs)
    outputs = Flatten()(outputs)
    outputs = Dense(units=256)(outputs)
    outputs = Activation('relu')(outputs)
    outputs = Dense(units=number_of_classification_labels)(outputs)

    outputs = Activation('softmax')(outputs)

    stnModel = Model(inputs=inputs, outputs=outputs)

    return(stnModel)
